# -*- coding: utf-8 -*-
"""Tabularizer transform, for pipelining."""
# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["mloning", "fkiraly"]
__all__ = ["Tabularizer"]

from sktime.datatypes import convert, convert_to
from sktime.transformations.base import BaseTransformer


class Tabularizer(BaseTransformer):
    """
    A transformer that turns time series/panel data into tabular data.

    This estimator converts nested pandas dataframe containing
    time-series/panel data with numpy arrays or pandas Series in
    dataframe cells into a tabular pandas dataframe with only primitives in
    cells. This is useful for transforming
    time-series/panel data into a format that is accepted by standard
    validation learning algorithms (as in sklearn).
    """

    _tags = {
        "fit_is_empty": True,
        "univariate-only": False,
        "scitype:transform-input": "Series",
        # what is the scitype of X: Series, or Panel
        "scitype:transform-output": "Primitives",
        # what is the scitype of y: None (not needed), Primitives, Series, Panel
        "scitype:instancewise": True,  # is this an instance-wise transform?
        "X_inner_mtype": ["nested_univ", "numpy3D"],
        # which mtypes do _fit/_predict support for X?
        "y_inner_mtype": "None",  # and for y?
    }

    def _transform(self, X, y=None):
        """Transform nested pandas dataframe into tabular dataframe.

        Parameters
        ----------
        X : pandas DataFrame or 3D np.ndarray
            panel of time series to transform
        y : ignored argument for interface compatibility

        Returns
        -------
        Xt : pandas DataFrame
            Transformed dataframe with only primitives in cells.
        """
        Xt = convert_to(X, to_type="numpyflat", as_scitype="Panel")
        return Xt

    def inverse_transform(self, X, y=None):
        """Transform tabular pandas dataframe into nested dataframe.

        Parameters
        ----------
        X : pandas DataFrame
            Tabular dataframe with primitives in cells.
        y : array-like, optional (default=None)

        Returns
        -------
        Xt : pandas DataFrame
            Transformed dataframe with series in cells.
        """
        Xt = convert(X, from_type="numpyflat", to_type="numpy3D", as_scitype="Panel")
        return Xt



# -*- coding: utf-8 -*-
"""Truncation transformer - truncate unequal length panels to lower/upper bounds."""
import numpy as np
import pandas as pd

from sktime.transformations.base import BaseTransformer

__all__ = ["TruncationTransformer"]
__author__ = ["abostrom"]


class TruncationTransformer(BaseTransformer):
    """Truncate unequal length panels to lower/upper bounds.

    Truncates all series in panel between lower/upper range bounds.

    Parameters
    ----------
    lower : int, optional (default=None) bottom range of the values to
                truncate can also be used to truncate to a specific length
                if None, will find the shortest sequence and use instead.
    upper : int, optional (default=None) upper range, only required when
                paired with lower.
                This is used to calculate the range between. exclusive.
                if None, will truncate from 0 to the lower bound.
    """

    _tags = {
        "scitype:transform-input": "Series",
        # what is the scitype of X: Series, or Panel
        "scitype:transform-output": "Series",
        # what scitype is returned: Primitives, Series, Panel
        "scitype:instancewise": False,  # is this an instance-wise transform?
        "X_inner_mtype": "nested_univ",  # which mtypes do _fit/_predict support for X?
        "y_inner_mtype": "None",  # which mtypes do _fit/_predict support for X?
        "fit_is_empty": False,  # is fit empty and can be skipped? Yes = True
        "capability:unequal_length:removes": True,
        # is transform result always guaranteed to be equal length (and series)?
    }

    def __init__(self, lower=None, upper=None):
        self.lower = lower
        self.upper = upper
        self.min_length = lower
        super(TruncationTransformer, self).__init__()

    @staticmethod
    def _get_min_length(X):
        def get_length(input):
            return min(map(lambda series: len(series), input))

        return min(map(get_length, X))

    def _fit(self, X, y=None):
        """Fit transformer to X and y.

        private _fit containing the core logic, called from fit

        Parameters
        ----------
        X : nested pandas DataFrame of shape [n_instances, n_features]
            each cell of X must contain pandas.Series
            Data to fit transform to
        y : ignored argument for interface compatibility
            Additional data, e.g., labels for transformation

        Returns
        -------
        self : reference to self
        """
        if self.lower is None:
            n_instances, _ = X.shape
            arr = [X.iloc[i, :].values for i in range(n_instances)]
            self.lower_ = self._get_min_length(arr)
        else:
            self.lower_ = self.lower

        return self

    def _transform(self, X, y=None):
        """Transform X and return a transformed version.

        private _transform containing core logic, called from transform

        Parameters
        ----------
        X : nested pandas DataFrame of shape [n_instances, n_features]
            each cell of X must contain pandas.Series
            Data to transform
        y : ignored argument for interface compatibility

        Returns
        -------
        Xt : nested pandas DataFrame of shape [n_instances, n_features]
            each cell of Xt contains pandas.Series
            transformed version of X
        """
        n_instances, _ = X.shape

        arr = [X.iloc[i, :].values for i in range(n_instances)]

        min_length = self._get_min_length(arr)

        if min_length < self.lower_:
            raise ValueError(
                "Error: min_length of series \
                    is less than the one found when fit or set."
            )

        # depending on inputs either find the shortest truncation.
        # or use the bounds.
        if self.upper is None:
            idxs = np.arange(self.lower_)
        else:
            idxs = np.arange(self.lower_, self.upper)

        truncate = [pd.Series([series.iloc[idxs] for series in out]) for out in arr]

        Xt = pd.DataFrame(truncate)

        return Xt

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.


        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        params = {"lower": 5}
        return params



# -*- coding: utf-8 -*-
"""Tests for RIFE."""

import numpy as np
import pandas as pd
import pytest
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import FunctionTransformer

from sktime.transformations.compose import FeatureUnion
from sktime.transformations.panel.compose import SeriesToPrimitivesRowTransformer
from sktime.transformations.panel.segment import RandomIntervalSegmenter
from sktime.transformations.panel.summarize import RandomIntervalFeatureExtractor
from sktime.utils._testing.panel import (
    _make_nested_from_array,
    make_classification_problem,
)
from sktime.utils.slope_and_trend import _slope


@pytest.mark.parametrize("n_instances", [1, 3])
@pytest.mark.parametrize("n_timepoints", [10, 20])
@pytest.mark.parametrize("n_intervals", [1, 3, "log", "sqrt", "random"])
@pytest.mark.parametrize(
    "features", [[np.mean], [np.mean, np.median], [np.mean, np.median, np.mean]]
)
def test_output_format_dim(n_instances, n_timepoints, n_intervals, features):
    """Test output format and dimensions."""
    X = _make_nested_from_array(
        np.ones(n_timepoints), n_instances=n_instances, n_columns=1
    )
    n_rows, n_cols = X.shape
    trans = RandomIntervalFeatureExtractor(n_intervals=n_intervals, features=features)
    Xt = trans.fit_transform(X)
    assert isinstance(Xt, pd.DataFrame)
    assert Xt.shape[0] == n_rows
    assert np.array_equal(Xt.values, np.ones(Xt.shape))


@pytest.mark.parametrize("bad_n_intervals", [0, "abc", 1.1, -1])
def test_bad_n_intervals(bad_n_intervals):
    """Check that exception is raised for bad input args."""
    X, y = make_classification_problem()
    with pytest.raises(ValueError):
        RandomIntervalFeatureExtractor(n_intervals=bad_n_intervals).fit(X)


@pytest.mark.parametrize(
    "bad_features", [0, "abc", {"a": 1}, (np.median, np.mean), [0, "abc"]]
)
def test_bad_features(bad_features):
    """Test that ValueError is raised for bad features."""
    X, y = make_classification_problem()
    with pytest.raises(ValueError):
        RandomIntervalFeatureExtractor(n_intervals=bad_features).fit(X)


@pytest.mark.parametrize("n_instances", [3, 5])
@pytest.mark.parametrize("n_timepoints", [10, 20])
@pytest.mark.parametrize("n_intervals", [1, 3, "log", "sqrt", "random"])
def test_results(n_instances, n_timepoints, n_intervals):
    """Check specific results."""
    X, _ = make_classification_problem(
        n_instances=n_instances, n_timepoints=n_timepoints, return_numpy=True
    )
    transformer = RandomIntervalFeatureExtractor(
        n_intervals=n_intervals, features=[np.mean, np.std]
    )
    Xt = transformer.fit_transform(X)
    Xt = Xt.loc[:, ~Xt.columns.duplicated()]
    # Check results
    intervals = transformer.intervals_
    for start, end in intervals:
        expected_mean = np.mean(X[:, 0, start:end], axis=-1)
        expected_std = np.std(X[:, 0, start:end], axis=-1)

        actual_means = Xt.loc[:, f"{start}_{end}_mean"].to_numpy().ravel()
        actual_stds = Xt.loc[:, f"{start}_{end}_std"].to_numpy().ravel()

        np.testing.assert_array_equal(actual_means, expected_mean)
        np.testing.assert_array_equal(actual_stds, expected_std)


def test_different_implementations():
    """Test against equivalent pipelines."""
    random_state = 1233
    X_train, y_train = make_classification_problem()

    # Compare with chained transformations.
    tran1 = RandomIntervalSegmenter(n_intervals=1, random_state=random_state)
    tran2 = SeriesToPrimitivesRowTransformer(
        FunctionTransformer(func=np.mean, validate=False), check_transformer=False
    )
    A = tran2.fit_transform(tran1.fit_transform(X_train))

    tran = RandomIntervalFeatureExtractor(
        n_intervals=1, features=[np.mean], random_state=random_state
    )
    B = tran.fit_transform(X_train)

    np.testing.assert_array_almost_equal(A, B)


@pytest.mark.xfail(reason="SeriesToPrimitivesTransformer will be deprecated, see 2179")
def test_different_pipelines():
    """Compare with transformer pipeline using TSFeatureUnion."""
    random_state = 1233
    X_train, y_train = make_classification_problem()
    steps = [
        (
            "segment",
            RandomIntervalSegmenter(n_intervals=1, random_state=random_state),
        ),
        (
            "transform",
            FeatureUnion(
                [
                    (
                        "mean",
                        SeriesToPrimitivesRowTransformer(
                            FunctionTransformer(func=np.mean, validate=False),
                            check_transformer=False,
                        ),
                    ),
                    (
                        "std",
                        SeriesToPrimitivesRowTransformer(
                            FunctionTransformer(func=np.std, validate=False),
                            check_transformer=False,
                        ),
                    ),
                    (
                        "slope",
                        SeriesToPrimitivesRowTransformer(
                            FunctionTransformer(func=_slope, validate=False),
                            check_transformer=False,
                        ),
                    ),
                ]
            ),
        ),
    ]
    pipe = Pipeline(steps)
    a = pipe.fit_transform(X_train)
    tran = RandomIntervalFeatureExtractor(
        n_intervals=1,
        features=[np.mean, np.std, _slope],
        random_state=random_state,
    )
    b = tran.fit_transform(X_train)
    np.testing.assert_array_equal(a, b)
    np.testing.assert_array_equal(pipe.steps[0][1].intervals_, tran.intervals_)



# -*- coding: utf-8 -*-
"""Test for distance path functionality."""
import numpy as np
import pytest

from sktime.distances._distance import _METRIC_INFOS, distance, distance_alignment_path
from sktime.distances.tests._utils import create_test_distance_numpy


def _validate_distance_alignment_path_result(
    x: np.ndarray,
    y: np.ndarray,
    metric_str: str,
    distance_path_callable,
    kwargs_dict: dict = None,
):
    if kwargs_dict is None:
        kwargs_dict = {}

    metric_str_result = distance_alignment_path(x, y, metric=metric_str, **kwargs_dict)
    assert isinstance(metric_str_result, tuple), (
        f"The result for a distance path using the string: {metric_str} as the "
        f'"metric" parameter should return a tuple. The return type provided is of '
        f"type {type(metric_str_result)}"
    )
    assert len(metric_str_result) == 2

    assert isinstance(metric_str_result[0], list), (
        f"The result for a distance path first return using the string: {metric_str} as"
        f" the metric parameter should return a tuple. The return type provided is of "
        f"type {type(metric_str_result)}"
    )

    assert isinstance(metric_str_result[1], float), (
        f"The result for a distance path first return using the string: {metric_str} as"
        f" the metric parameter should return a tuple. The return type provided is of "
        f"type {type(metric_str_result)}"
    )

    distance_result = distance(x, y, metric=metric_str)
    path_result = distance_path_callable(x, y)

    assert distance_result == metric_str_result[1]
    assert distance_result == path_result[1]

    metric_str_result_cm = distance_alignment_path(
        x, y, metric=metric_str, return_cost_matrix=True, **kwargs_dict
    )

    assert isinstance(metric_str_result_cm, tuple), (
        f"The result for a distance path using the string: {metric_str} as the "
        f'"metric" parameter should return a tuple. The return type provided is of '
        f"type {type(metric_str_result)}"
    )
    assert len(metric_str_result_cm) == 3

    assert isinstance(metric_str_result_cm[0], list), (
        f"The result for a distance path first return using the string: {metric_str} as"
        f" the metric parameter should return a tuple. The return type provided is of "
        f"type {type(metric_str_result_cm[0])}"
    )

    assert isinstance(metric_str_result_cm[1], float), (
        f"The result for a distance path first return using the string: {metric_str} as"
        f" the metric parameter should return a tuple. The return type provided is of "
        f"type {type(metric_str_result)}"
    )
    assert metric_str_result_cm[0] == metric_str_result[0]
    assert metric_str_result_cm[1] == metric_str_result[1]


@pytest.mark.parametrize("dist", _METRIC_INFOS)
def test_distance_alignment_path(dist):
    """Test the distance paths."""
    if dist.dist_alignment_path_func is not None:

        x = create_test_distance_numpy(10, 1)
        y = create_test_distance_numpy(10, 1, random_state=2)
        _validate_distance_alignment_path_result(
            x=x,
            y=y,
            metric_str=dist.canonical_name,
            distance_path_callable=dist.dist_alignment_path_func,
        )



#!/usr/bin/env python3 -u
# -*- coding: utf-8 -*-
# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)
"""Transformer to impute missing values in series."""

__author__ = ["aiwalter"]
__all__ = ["Imputer"]


from warnings import warn

import numpy as np
from sklearn.utils import check_random_state

from sktime.forecasting.base import ForecastingHorizon
from sktime.forecasting.trend import PolynomialTrendForecaster
from sktime.transformations.base import BaseTransformer


class Imputer(BaseTransformer):
    """Missing value imputation.

    The Imputer transforms input series by replacing missing values according
    to an imputation strategy specified by `method`.

    Parameters
    ----------
    method : str, default="drift"
        Method to fill the missing values values.

        * "drift" : drift/trend values by sktime.PolynomialTrendForecaster(degree=1)
            first, X in transform() is filled with ffill then bfill
            then PolynomialTrendForecaster(degree=1) is fitted to filled X, and
            predict values are queried at indices which had missing values
        * "linear" : linear interpolation, uses pd.Series.interpolate()
            WARNING: This method can not extrapolate, so it is fitted always on the
            data given to transform().
        * "nearest" : use nearest value, uses pd.Series.interpolate()
        * "constant" : same constant value (given in arg value) for all NaN
        * "mean" : pd.Series.mean() of *fit* data
        * "median" : pd.Series.median() of *fit* data
        * "backfill" ot "bfill" : adapted from pd.Series.fillna()
        * "pad" or "ffill" : adapted from pd.Series.fillna()
        * "random" : random values between pd.Series.min() and .max() of *fit* data
            if pd.Series dtype is int, sample is uniform discrete
            if pd.Series dtype is float, sample is uniform continuous
        * "forecaster" : use an sktime Forecaster, given in param forecaster.
            First, X in *fit* is filled with ffill then bfill
            then forecaster is fitted to filled X, and *predict* values are queried
            at indices of X data in *transform* which had missing values
        For the following methods, the train data is used to fit them:
        "drift", "mean", "median", "random". For all other methods, the
        transform data is sufficient to compute the impute values.

    missing_values : int/float/str, default=None
        The placeholder for the missing values. All occurrences of
        missing_values will be imputed, in addition to np.nan.
        If None, then only np.nan values are imputed.
    value : int/float, default=None
        Value to use to fill missing values when method="constant".
    forecaster : Any Forecaster based on sktime.BaseForecaster, default=None
        Use a given Forecaster to impute by insample predictions when
        method="forecaster". Before fitting, missing data is imputed with
        method="ffill" or "bfill" as heuristic. in case of multivariate X,
        the forecaster is applied separete to each column like a
        ColumnEnsembleForecaster.
    random_state : int/float/str, optional
        Value to set random.seed() if method="random", default None

    Examples
    --------
    >>> from sktime.transformations.series.impute import Imputer
    >>> from sktime.datasets import load_airline
    >>> from sktime.forecasting.model_selection import temporal_train_test_split
    >>> y = load_airline()
    >>> y_train, y_test = temporal_train_test_split(y)
    >>> transformer = Imputer(method="drift")
    >>> transformer.fit(y_train)
    Imputer(...)
    >>> y_test.iloc[3] = np.nan
    >>> y_hat = transformer.transform(y_test)
    """

    _tags = {
        "scitype:transform-input": "Series",
        # what is the scitype of X: Series, or Panel
        "scitype:transform-output": "Series",
        # what scitype is returned: Primitives, Series, Panel
        "scitype:instancewise": True,  # is this an instance-wise transform?
        "X_inner_mtype": ["pd.DataFrame"],
        # which mtypes do _fit/_predict support for X?
        "y_inner_mtype": "None",  # which mtypes do _fit/_predict support for y?
        "fit_is_empty": False,
        "handles-missing-data": True,
        "skip-inverse-transform": True,
        "univariate-only": False,
        "capability:missing_values:removes": True,
        # is transform result always guaranteed to contain no missing values?
    }

    def __init__(
        self,
        method="drift",
        random_state=None,
        value=None,
        forecaster=None,
        missing_values=None,
    ):

        self.method = method
        self.missing_values = missing_values
        self.value = value
        self.forecaster = forecaster
        self.random_state = random_state
        super(Imputer, self).__init__()

    def _fit(self, X, y=None):
        """Fit transformer to X and y.

        private _fit containing the core logic, called from fit

        Parameters
        ----------
        X : Series or Panel of mtype X_inner_mtype
            if X_inner_mtype is list, _fit must support all types in it
            Data to fit transform to
        y : Series or Panel of mtype y_inner_mtype, default=None
            Additional data, e.g., labels for transformation

        Returns
        -------
        self: reference to self
        """
        self._check_method()
        # all methods of Imputer that are actually doing a fit are
        # implemented here. Some methods dont need fit, so they are just
        # impleented in _transform
        if self.method in ["drift", "forecaster"]:
            # save train data as needed for multivariate fitting int _fit()
            self._X = X.copy()
            self._y = y.copy() if y is not None else None
            if self.method == "drift":
                self._forecaster = PolynomialTrendForecaster(degree=1)
            elif self.method == "forecaster":
                self._forecaster = self.forecaster.clone()
        elif self.method == "mean":
            self._mean = X.mean()
        elif self.method == "median":
            self._median = X.median()
        elif self.method == "random":
            # save train data to get min() and max() in transform() for each column
            self._X = X.copy()

    def _transform(self, X, y=None):
        """Transform X and return a transformed version.

        private _transform containing the core logic, called from transform

        Parameters
        ----------
        X : pd.Series or pd.DataFrame
            Data to be transformed
        y : ignored argument for interface compatibility
            Additional data, e.g., labels for transformation

        Returns
        -------
        X : pd.Series or pd.DataFrame, same type as X
            transformed version of X
        """
        X = X.copy()

        # TODO v0.13.0: Remove this if statement and warning
        if self.method in ["drift", "mean", "median", "random"]:
            warn(
                """Imputer methods "drift", "mean", "median", "random" have been
                moved to the fit() method, so usage via transform() is
                deprecated. To still fit on the transform data only, please use the
                new FitInTransform transformer.
                """,
                DeprecationWarning,
            )
        # replace missing_values with np.nan
        if self.missing_values:
            X = X.replace(to_replace=self.missing_values, value=np.nan)

        if not _has_missing_values(X):
            return X

        if self.method == "random":
            for col in X.columns:
                X[col] = X[col].apply(
                    lambda i: self._get_random(col) if np.isnan(i) else i
                )
        elif self.method == "constant":
            X = X.fillna(value=self.value)
        elif self.method in ["backfill", "bfill", "pad", "ffill"]:
            X = X.fillna(method=self.method)
        elif self.method == "drift":
            X = self._impute_with_forecaster(X, y)
        elif self.method == "forecaster":
            X = self._impute_with_forecaster(X, y)
        elif self.method == "mean":
            X = X.fillna(value=self._mean)
        elif self.method == "median":
            X = X.fillna(value=self._median)
        elif self.method in ["nearest", "linear"]:
            X = X.interpolate(method=self.method)
        else:
            raise ValueError(f"`method`: {self.method} not available.")
        # fill first/last elements of series,
        # as some methods (e.g. "linear") cant impute those
        X = X.fillna(method="ffill").fillna(method="backfill")
        return X

    def _check_method(self):
        method = self.method
        if method not in [
            "mean",
            "drift",
            "linear",
            "nearest",
            "constant",
            "median",
            "backfill",
            "bfill",
            "pad",
            "ffill",
            "random",
            "forecaster",
        ]:
            raise ValueError(f"Given method {method} is not an allowed method.")
        if (
            self.value is not None
            and method != "constant"
            or method == "constant"
            and self.value is None
        ):
            raise ValueError(
                """Imputing with a value can only be
                used if method="constant" and if parameter "value" is not None"""
            )
        elif (
            self.forecaster is not None
            and method != "forecaster"
            or method == "forecaster"
            and self.forecaster is None
        ):
            raise ValueError(
                """Imputing with a forecaster can only be used if
                method=\"forecaster\" and if arg forecaster is not None"""
            )
        else:
            pass

    def _get_random(self, col):
        """Create a random int or float value.

        Parameters
        ----------
        col : str
            Column name

        Returns
        -------
        int/float
            Random int or float between min and max of X
        """
        rng = check_random_state(self.random_state)
        # check if series contains only int or int-like values (e.g. 3.0)
        if (self._X[col].dropna() % 1 == 0).all():
            return rng.randint(self._X[col].min(), self._X[col].max())
        else:
            return rng.uniform(self._X[col].min(), self._X[col].max())

    def _impute_with_forecaster(self, X, y):
        """Use a given forecaster for imputation by in-sample predictions.

        Parameters
        ----------
        X : pd.DataFrame
            Series to impute.
        y : pd.DataFrame
            Exog data for forecaster.

        Returns
        -------
        Xt : pd.DataFrame
            Series with imputed values.
        """
        for col in X.columns:
            if _has_missing_values(X[col]):
                # define fh based on index of missing values
                na_index = X[col].index[X[col].isna()]
                fh = ForecastingHorizon(values=na_index, is_relative=False)

                # fill NaN before fitting with ffill and backfill (heuristic)

                self._forecaster.fit(
                    y=self._X[col].fillna(method="ffill").fillna(method="backfill"),
                    X=self._y[col].fillna(method="ffill").fillna(method="backfill")
                    if self._y is not None
                    else None,
                )

                # replace missing values with predicted values
                X[col][na_index] = self._forecaster.predict(fh=fh, X=y)
        return X

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.


        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        from sktime.forecasting.exp_smoothing import ExponentialSmoothing

        return [
            {"method": "drift"},
            {"method": "linear"},
            {"method": "nearest"},
            {"method": "constant", "value": 1},
            {"method": "median"},
            {"method": "backfill"},
            {"method": "bfill"},
            {"method": "pad"},
            {"method": "random"},
            {"method": "forecaster", "forecaster": ExponentialSmoothing()},
        ]


def _has_missing_values(X):
    return X.isnull().to_numpy().any()



# -*- coding: utf-8 -*-

__author__ = ["mloning"]
__all__ = ["ESTIMATOR_TEST_PARAMS", "EXCLUDE_ESTIMATORS", "EXCLUDED_TESTS"]

import numpy as np
from sklearn.preprocessing import FunctionTransformer, StandardScaler

from sktime.annotation.clasp import ClaSPSegmentation
from sktime.base import BaseEstimator
from sktime.forecasting.exp_smoothing import ExponentialSmoothing
from sktime.forecasting.structural import UnobservedComponents
from sktime.registry import (
    BASE_CLASS_LIST,
    BASE_CLASS_LOOKUP,
    ESTIMATOR_TAG_LIST,
    TRANSFORMER_MIXIN_LIST,
)
from sktime.regression.compose import ComposableTimeSeriesForestRegressor
from sktime.transformations.base import BaseTransformer
from sktime.transformations.panel.compose import (
    ColumnTransformer,
    SeriesToPrimitivesRowTransformer,
    SeriesToSeriesRowTransformer,
)
from sktime.transformations.panel.random_intervals import RandomIntervals
from sktime.transformations.panel.shapelet_transform import RandomShapeletTransform
from sktime.transformations.panel.summarize import FittedParamExtractor

# The following estimators currently do not pass all unit tests
# https://github.com/alan-turing-institute/sktime/issues/1627
EXCLUDE_ESTIMATORS = [
    # ConditionalDeseasonalizer and STLtransformer still need refactoring
    #  (see PR 1773, blocked through open discussion) escaping until then
    "ConditionalDeseasonalizer",
    "STLTransformer",
    # SFA is non-compliant with any transformer interfaces, #2064
    "SFA",
    # requires y in fit, this is incompatible with the old testing framework
    #    unless it inherits from the old mixins, which hard coded the y
    #    should be removed once test_all_transformers has been refactored to scenarios
    "TSFreshRelevantFeatureExtractor",
    # PlateauFinder seems to be broken, see #2259
    "PlateauFinder",
]


EXCLUDED_TESTS = {
    # known issue when X is passed, wrong time indices are returned, #1364
    "StackingForecaster": ["test_predict_time_index_with_X"],
    # known side effects on multivariate arguments, #2072
    "WindowSummarizer": ["test_methods_have_no_side_effects"],
    # test fails in the Panel case for Differencer, see #2522
    "Differencer": ["test_transform_inverse_transform_equivalent"],
    # tagged in issue #2490
    "SignatureClassifier": [
        "test_classifier_on_unit_test_data",
        "test_classifier_on_basic_motions",
    ],
    # test fail with deep problem with pickling inside tensorflow.
    "CNNClassifier": [
        "test_fit_idempotent",
        "test_persistence_via_pickle",
    ],
    # pickling problem with local method see #2490
    "ProximityStump": [
        "test_persistence_via_pickle",
        "test_fit_does_not_overwrite_hyper_params",
    ],
    "ProximityTree": [
        "test_persistence_via_pickle",
        "test_fit_does_not_overwrite_hyper_params",
    ],
    "ProximityForest": [
        "test_persistence_via_pickle",
        "test_fit_does_not_overwrite_hyper_params",
    ],
    # sth is not quite right with the RowTransformer-s changing state,
    #   but these are anyway on their path to deprecation, see #2370
    "SeriesToPrimitivesRowTransformer": ["test_methods_do_not_change_state"],
    "SeriesToSeriesRowTransformer": ["test_methods_do_not_change_state"],
    # ColumnTransformer still needs to be refactored, see #2537
    "ColumnTransformer": ["test_methods_do_not_change_state"],
    "ForecastingGridSearchCV": ["test_score"],  # unknown root cause, see #2751
    "ForecastingRandomizedSearchCV": ["test_score"],  # unknown root cause, see #2751
    # failure of test_score came up after a change of metric default params
    # there should be no such failure, and all other algorithms pass. #2751 to track
}

# We here configure estimators for basic unit testing, including setting of
# required hyper-parameters and setting of hyper-parameters for faster training.
SERIES_TO_SERIES_TRANSFORMER = StandardScaler()
SERIES_TO_PRIMITIVES_TRANSFORMER = FunctionTransformer(
    np.mean, kw_args={"axis": 0}, check_inverse=False
)
TRANSFORMERS = [
    (
        "transformer1",
        SeriesToSeriesRowTransformer(
            SERIES_TO_SERIES_TRANSFORMER, check_transformer=False
        ),
    ),
    (
        "transformer2",
        SeriesToSeriesRowTransformer(
            SERIES_TO_SERIES_TRANSFORMER, check_transformer=False
        ),
    ),
]
ESTIMATOR_TEST_PARAMS = {
    FittedParamExtractor: {
        "forecaster": ExponentialSmoothing(),
        "param_names": ["initial_level"],
    },
    SeriesToPrimitivesRowTransformer: {
        "transformer": SERIES_TO_PRIMITIVES_TRANSFORMER,
        "check_transformer": False,
    },
    SeriesToSeriesRowTransformer: {
        "transformer": SERIES_TO_SERIES_TRANSFORMER,
        "check_transformer": False,
    },
    ColumnTransformer: {
        "transformers": [(name, estimator, [0]) for name, estimator in TRANSFORMERS]
    },
    RandomShapeletTransform: {
        "max_shapelets": 5,
        "n_shapelet_samples": 50,
        "batch_size": 20,
    },
    RandomIntervals: {
        "n_intervals": 3,
    },
    ComposableTimeSeriesForestRegressor: {"n_estimators": 3},
    UnobservedComponents: {"level": "local level"},
    ClaSPSegmentation: {"period_length": 5, "n_cps": 1},
}

# We use estimator tags in addition to class hierarchies to further distinguish
# estimators into different categories. This is useful for defining and running
# common tests for estimators with the same tags.
VALID_ESTIMATOR_TAGS = tuple(ESTIMATOR_TAG_LIST)

# These methods should not change the state of the estimator, that is, they should
# not change fitted parameters or hyper-parameters. They are also the methods that
# "apply" the fitted estimator to data and useful for checking results.
NON_STATE_CHANGING_METHODS = (
    "predict",
    "predict_var",
    "predict_proba",
    "decision_function",
    "transform",
    # todo: add this back
    # escaping this, since for some estimators
    #   the input format of inverse_transform assumes special col names
    # "inverse_transform",
)

# The following gives a list of valid estimator base classes.
VALID_TRANSFORMER_TYPES = tuple(TRANSFORMER_MIXIN_LIST) + (BaseTransformer,)

VALID_ESTIMATOR_BASE_TYPES = tuple(BASE_CLASS_LIST)

VALID_ESTIMATOR_TYPES = (
    BaseEstimator,
    *VALID_ESTIMATOR_BASE_TYPES,
    *VALID_TRANSFORMER_TYPES,
)

VALID_ESTIMATOR_BASE_TYPE_LOOKUP = BASE_CLASS_LOOKUP



# -*- coding: utf-8 -*-
# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)
"""Machine type converters for scitypes.

Exports
-------
convert_to(obj, to_type: str, as_scitype: str, store=None)
    converts object "obj" to type "to_type", considerd as "as_scitype"

convert(obj, from_type: str, to_type: str, as_scitype: str, store=None)
    same as convert_to, without automatic identification of "from_type"

mtype(obj, as_scitype: str)
    returns "from_type" of obj, considered as "as_scitype"
---

Function signature of convert

Parameters
----------
obj : object to convert - any type, should comply with mtype spec for as_scitype
from_type : str - the type to convert "obj" to, a valid mtype string
to_type : str - the type to convert "obj" to, a valid mtype string
as_scitype : str - name of scitype the object "obj" is considered as
store : reference of storage for lossy conversions, default=None (no store)

Returns
-------
converted_obj : to_type - object obj converted to to_type

---

Function signature of convert_to

Parameters
----------
obj : object to convert - any type, should comply with mtype spec for as_scitype
to_type : str - the type to convert "obj" to, a valid mtype string
as_scitype : str - name of scitype the object "obj" is considered as
store : reference of storage for lossy conversions, default=None (no store)

Returns
-------
converted_obj : to_type - object obj converted to to_type

---

Function signature of mtype

Parameters
----------
obj : object to convert - any type, should comply with mtype spec for as_scitype
as_scitype : str - name of scitype the object "obj" is considered as

Returns
-------
str - the type to convert "obj" to, a valid mtype string
    or None, if obj is None
"""

__author__ = ["fkiraly"]

__all__ = [
    "convert",
    "convert_to",
]

from copy import deepcopy

import numpy as np
import pandas as pd

from sktime.datatypes._check import mtype as infer_mtype
from sktime.datatypes._hierarchical import convert_dict_Hierarchical
from sktime.datatypes._panel import convert_dict_Panel
from sktime.datatypes._proba import convert_dict_Proba
from sktime.datatypes._registry import mtype_to_scitype
from sktime.datatypes._series import convert_dict_Series
from sktime.datatypes._table import convert_dict_Table

# pool convert_dict-s and infer_mtype_dict-s
convert_dict = dict()
convert_dict.update(convert_dict_Series)
convert_dict.update(convert_dict_Panel)
convert_dict.update(convert_dict_Hierarchical)
convert_dict.update(convert_dict_Table)
convert_dict.update(convert_dict_Proba)


def convert(
    obj,
    from_type: str,
    to_type: str,
    as_scitype: str = None,
    store=None,
    store_behaviour: str = None,
):
    """Convert objects between different machine representations, subject to scitype.

    Parameters
    ----------
    obj : object to convert - any type, should comply with mtype spec for as_scitype
    from_type : str - the type to convert "obj" to, a valid mtype string
        valid mtype strings, with explanation, are in datatypes.MTYPE_REGISTER
    to_type : str - the type to convert "obj" to, a valid mtype string
        valid mtype strings, with explanation, are in datatypes.MTYPE_REGISTER
    as_scitype : str, optional - name of scitype the object "obj" is considered as
        default = inferred from from_type
        valid scitype strings, with explanation, are in datatypes.SCITYPE_REGISTER
    store : optional, reference of storage for lossy conversions, default=None (no ref)
        is updated by side effect if not None and store_behaviour="reset" or "update"
    store_behaviour : str, optional, one of None (default), "reset", "freeze", "update"
        "reset" - store is emptied and then updated from conversion
        "freeze" - store is read-only, may be read/used by conversion but not changed
        "update" - store is updated from conversion and retains previous contents
        None - automatic: "update" if store is empty and not None; "freeze", otherwise

    Returns
    -------
    converted_obj : to_type - object obj converted to to_type
                    if obj was None, returns None

    Raises
    ------
    KeyError if conversion is not implemented
    TypeError or ValueError if inputs do not match specification
    """
    if obj is None:
        return None

    # input type checks
    if not isinstance(to_type, str):
        raise TypeError("to_type must be a str")
    if not isinstance(from_type, str):
        raise TypeError("from_type must be a str")
    if as_scitype is None:
        as_scitype = mtype_to_scitype(to_type)
    elif not isinstance(as_scitype, str):
        raise TypeError("as_scitype must be str or None")
    if store is not None and not isinstance(store, dict):
        raise TypeError("store must be a dict or None")
    if store_behaviour not in [None, "reset", "freeze", "update"]:
        raise ValueError(
            'store_behaviour must be one of "reset", "freeze", "update", or None'
        )
    if store_behaviour is None and store == {}:
        store_behaviour = "update"
    if store_behaviour is None and store != {}:
        store_behaviour = "freeze"

    key = (from_type, to_type, as_scitype)

    if key not in convert_dict.keys():
        raise NotImplementedError(
            "no conversion defined from type " + str(from_type) + " to " + str(to_type)
        )

    if store_behaviour == "freeze":
        store = deepcopy(store)
    elif store_behaviour == "reset":
        # note: this is a side effect on store
        store.clear()
    elif store_behaviour == "update":
        # store is passed to convert_obj by reference, unchanged
        # this "elif" is here for clarity, to cover all three values
        pass
    else:
        raise RuntimeError(
            "bug: unrechable condition error, store_behaviour has unexpected value"
        )

    converted_obj = convert_dict[key](obj, store=store)

    return converted_obj


# conversion based on queriable type to specified target
def convert_to(
    obj,
    to_type: str,
    as_scitype: str = None,
    store=None,
    store_behaviour: str = None,
):
    """Convert object to a different machine representation, subject to scitype.

    Parameters
    ----------
    obj : object to convert - any type, should comply with mtype spec for as_scitype
    to_type : str - the type to convert "obj" to, a valid mtype string
            or list of str, this specifies admissible types for conversion to
        valid mtype strings, with explanation, are in datatypes.MTYPE_REGISTER
    as_scitype : str, optional - name of scitype the object "obj" is considered as
        pre-specifying the scitype reduces the number of checks done in type inference
        valid scitype strings, with explanation, are in datatypes.SCITYPE_REGISTER
        default = inferred from mtype of obj, which is in turn inferred internally
    store : reference of storage for lossy conversions, default=None (no store)
        is updated by side effect if not None and store_behaviour="reset" or "update"
    store_behaviour : str, optional, one of None (default), "reset", "freeze", "update"
        "reset" - store is emptied and then updated from conversion
        "freeze" - store is read-only, may be read/used by conversion but not changed
        "update" - store is updated from conversion and retains previous contents
        None - automatic: "update" if store is empty and not None; "freeze", otherwise

    Returns
    -------
    converted_obj : to_type - object obj, or obj converted to target mtype as follows:
        case 1: mtype of obj is equal to to_type, or a list element of to_type
            no conversion happens, converted_obj = obj
        case 2: to_type is a str, and not equal to mtype of obj
            converted_obj is obj converted to to_type
        case 3: to_type is list of str, and mtype of obj is not in that list
            converted_obj is converted to the first mtype in to_type
                that is of same scitype as obj
        case 4: if obj was None, converted_obj is also None

    Raises
    ------
    TypeError if machine type of input "obj" is not recognized
    TypeError if to_type contains no mtype compatible with mtype of obj
    KeyError if conversion that would be conducted is not implemented
    TypeError or ValueError if inputs do not match specification
    """
    if obj is None:
        return None

    # input checks on to_type, as_scitype; coerce to_type, as_scitype to lists
    to_type = _check_str_or_list_of_str(to_type, obj_name="to_type")

    # sub-set a preliminary set of as_scitype from to_type, as_scitype
    if as_scitype is not None:
        # if not None, subset to types compatible between to_type and as_scitype
        as_scitype = _check_str_or_list_of_str(as_scitype, obj_name="as_scitype")
        potential_scitypes = mtype_to_scitype(to_type)
        as_scitype = list(set(potential_scitypes).intersection(as_scitype))
    else:
        # if None, infer from to_type
        as_scitype = mtype_to_scitype(to_type)

    # now further narrow down as_scitype by inference from the obj
    from_type = infer_mtype(obj=obj, as_scitype=as_scitype)
    as_scitype = mtype_to_scitype(from_type)

    # if to_type is a list, we do the following:
    # if on the list, then don't do a conversion (convert to from_type)
    # if not on the list, we find and convert to first mtype that has same scitype
    if isinstance(to_type, list):
        # no conversion of from_type is in the list
        if from_type in to_type:
            to_type = from_type
        # otherwise convert to first element of same scitype
        else:
            same_scitype_mtypes = [
                mtype for mtype in to_type if mtype_to_scitype(mtype) == as_scitype
            ]
            if len(same_scitype_mtypes) == 0:
                raise TypeError(
                    "to_type contains no mtype compatible with the scitype of obj,"
                    f"which is {as_scitype}"
                )
            to_type = same_scitype_mtypes[0]

    converted_obj = convert(
        obj=obj,
        from_type=from_type,
        to_type=to_type,
        as_scitype=as_scitype,
        store=store,
        store_behaviour=store_behaviour,
    )

    return converted_obj


def _conversions_defined(scitype: str):
    """Return an indicator matrix which conversions are defined for scitype.

    Parameters
    ----------
    scitype: str - name of scitype for which conversions are queried
        valid scitype strings, with explanation, are in datatypes.SCITYPE_REGISTER

    Returns
    -------
    conv_df: pd.DataFrame, columns and index is list of mtypes for scitype
            entry of row i, col j is 1 if conversion from i to j is defined,
                                     0 if conversion from i to j is not defined
    """
    pairs = [(x[0], x[1]) for x in list(convert_dict.keys()) if x[2] == scitype]
    cols0 = set([x[0] for x in list(convert_dict.keys()) if x[2] == scitype])
    cols1 = set([x[1] for x in list(convert_dict.keys()) if x[2] == scitype])
    cols = sorted(list(cols0.union(cols1)))

    mat = np.zeros((len(cols), len(cols)), dtype=int)
    nkeys = len(cols)
    for i in range(nkeys):
        for j in range(nkeys):
            if (cols[i], cols[j]) in pairs:
                mat[i, j] = 1

    conv_df = pd.DataFrame(mat, index=cols, columns=cols)

    return conv_df


def _check_str_or_list_of_str(obj, obj_name="obj"):
    """Check whether obj is str or list of str; coerces to list of str."""
    if isinstance(obj, list):
        if not np.all(isinstance(x, str) for x in obj):
            raise TypeError(f"{obj} must be a str or list of str")
        else:
            return obj
    elif not isinstance(obj, str):
        raise TypeError(f"{obj} must be a str or list of str")
    else:
        return [obj]



# -*- coding: utf-8 -*-
# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)
"""Kalman Filter transformers unit tests."""

__author__ = ["NoaBenAmi"]

import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal

from sktime.transformations.series.kalman_filter import (
    KalmanFilterTransformerFP,
    KalmanFilterTransformerPK,
)

# ts stands for time steps
ts = 10


def create_data(shape, missing_values=False, p=0.15, mult=10):
    """Create random ndarray of shape `shape`.

    The result array will contain missing values (represented by np.nan)
    if parameter `missing_values` is set to true.
    """
    if isinstance(shape, int):
        shape = (shape,)
    data = np.random.rand(*shape) * mult

    if missing_values:
        time_steps = shape[0]
        measurement_dim = shape[1]
        pr = max(int(time_steps * p), 4)
        for t in range(time_steps):
            if t % pr == 0:
                data[t] = [np.nan] * measurement_dim
    return data


def rand_list(shape, length=ts, mult=10):
    """Return list with `length` random ndarrays of shape `shape`."""
    return [create_data(shape, mult=mult) for _ in range(length)]


# state_dim = 3, measurement_dim = 3, time_steps = ts (10)
params_3_3_dynamic = {
    "process_noise": create_data((ts, 3, 3)),  # random ndarray of shape (ts, 3, 3),
    "measurement_noise": rand_list(
        (3, 3), length=ts
    ),  # [`ts` random ndarrays of shape (3, 3)],
    "measurement_function": create_data((3, 3)),  # random ndarray of shape (3, 3),
    "initial_state_covariance": create_data((3, 3)),  # random ndarray of shape (3, 3),
}

# state_dim = 3, measurement_dim = 3, time_steps = ts (10)
params_3_3_static = {
    "state_transition": create_data((3, 3)),  # random ndarray of shape (3,3),
    "process_noise": create_data((3, 3)),  # random ndarray of shape (3,3),
    "measurement_noise": create_data((3, 3)),  # random ndarray of shape (3,3),
    "initial_state": create_data(3),  # random ndarray of shape (3,),
    "initial_state_covariance": create_data((3, 3)),  # random ndarray of shape (3,3)
}

# state_dim = 2, measurement_dim = 3, time_steps = ts (10)
params_2_3_ = {
    "state_transition": rand_list(
        (2, 2), length=ts
    ),  # [`ts` random ndarrays of shape (2,2)],
    "measurement_noise": create_data((3, 3)),  # random ndarray of shape (3, 3),
    "measurement_function": rand_list(
        (3, 2), length=ts
    ),  # [`ts` random ndarrays of shape (3,2)],
    "initial_state": create_data(2),  # random ndarray of shape (2,),
    "initial_state_covariance": create_data((2, 2)),  # random ndarray of shape (2,2)
}

# state_dim = 1, measurement_dim = 1, time_steps = ts (10)
params_1_1_arrays = {
    "state_transition": create_data((ts, 1, 1)),  # random ndarray of shape (ts, 1, 1)
    "process_noise": create_data((ts, 1, 1)),  # random ndarray of shape (ts, 1, 1)
    "initial_state": create_data(1),  # random ndarray of shape (1,)
    "initial_state_covariance": create_data((1, 1)),  # random ndarray of shape (1, 1)
}

# state_dim = 1, measurement_dim = 1, time_steps = ts (10)
params_1_1_lists = {
    "process_noise": rand_list(
        (1, 1), length=ts
    ),  # [`ts` random ndarrays of shape (1,1)],
    "measurement_noise": rand_list(
        (1, 1), length=ts
    ),  # [`ts` random ndarrays of shape (1,1)],
    "measurement_function": rand_list(
        (1, 1), length=ts
    ),  # [`ts` random ndarrays of shape (1,1)],
    "initial_state": create_data(1),  # random ndarray of shape (1,)
}

# state_dim = 3, measurement_dim = 1, time_steps = ts (10)
params_3_1_lists = {
    "state_transition": rand_list(
        (3, 3), length=ts
    ),  # [`ts` random ndarrays of shape (3,3)],
    "process_noise": rand_list(
        (3, 3), length=ts
    ),  # [`ts` random ndarrays of shape (3,3)],
    "measurement_noise": rand_list(
        (1, 1), length=ts
    ),  # [`ts` random ndarrays of shape (1,1)],
    "measurement_function": rand_list(
        (1, 3), length=ts
    ),  # [`ts` random ndarrays of shape (1,3)],
}


def get_params_mapping(params):
    """Transform parameters names.

    From KalmanFilterTransformerPK, KalmanFilterTransformerFP naming
    forms to `pykalman`'s naming form.
    """
    params_mapping = {
        "state_transition": "transition_matrices",
        "process_noise": "transition_covariance",
        "measurement_offsets": "observation_offsets",
        "transition_offsets": "transition_offsets",
        "measurement_noise": "observation_covariance",
        "measurement_function": "observation_matrices",
        "initial_state": "initial_state_mean",
        "initial_state_covariance": "initial_state_covariance",
    }
    if params is None:
        return None
    if isinstance(params, str) and params == "all":
        return list(params_mapping.values())
    return [params_mapping[param] for param in params]


def init_kf_pykalman(
    state_dim,
    measurement_dim=None,
    state_transition=None,
    process_noise=None,
    measurement_noise=None,
    measurement_function=None,
    initial_state=None,
    initial_state_covariance=None,
    transition_offsets=None,
    measurement_offsets=None,
    estimate_matrices=None,
    denoising=False,
):
    """Initiate instance of `pykalman`'s `KalmanFilter`."""
    from pykalman.standard import KalmanFilter

    em_vars = get_params_mapping(params=estimate_matrices)
    kf_pykalman = KalmanFilter(
        transition_matrices=state_transition,
        observation_matrices=measurement_function,
        transition_covariance=process_noise,
        observation_covariance=measurement_noise,
        transition_offsets=transition_offsets,
        observation_offsets=measurement_offsets,
        initial_state_mean=initial_state,
        initial_state_covariance=initial_state_covariance,
        n_dim_state=state_dim,
        n_dim_obs=measurement_dim,
        em_vars=em_vars,
    )
    return kf_pykalman


def init_kf_filterpy(measurements, adapter, n=10, y=None):
    """Adjust params and measurements.

    Given measurements and adapter, adjust params and measurements to
    `FilterPy` usable form.
    """
    y_dim = 1 if y is None else y.shape[-1]

    G = (
        np.eye(adapter.state_dim, y_dim)
        if adapter.control_transition is None
        else np.atleast_2d(adapter.control_transition)
    )

    matrices = {
        "Fs": [adapter.F_] * n if adapter.F_.ndim == 2 else [f for f in adapter.F_],
        "Qs": [adapter.Q_] * n if adapter.Q_.ndim == 2 else [q for q in adapter.Q_],
        "Rs": [adapter.R_] * n if adapter.R_.ndim == 2 else [r for r in adapter.R_],
        "Hs": [adapter.H_] * n if adapter.H_.ndim == 2 else [h for h in adapter.H_],
        "Bs": [G] * n if G.ndim == 2 else [g for g in G],
        "us": None if y is None else ([y] * n if y.ndim == 1 else [_y for _y in y]),
        "x": adapter.X0_,
        "P": adapter.P0_,
    }
    data = [None if any(np.isnan(d)) else d.copy() for d in measurements]

    return matrices, data


@pytest.mark.parametrize(
    "params, measurements",
    [  # test case 1 -
        #   state_dim = 3, measurement_dim = 3, params - params_3_3_dynamic
        (dict(params_3_3_dynamic, state_dim=3), create_data((ts, 3))),
        # H and X0 will be estimated using em algorithm.
        (
            dict(
                params_3_3_dynamic,
                state_dim=3,
                estimate_matrices=["measurement_function", "initial_state"],
            ),
            create_data((ts, 3)),
        ),
        # test case 2 -
        #   state_dim = 3, measurement_dim = 3, params - params_3_3_static
        (dict(params_3_3_static, state_dim=3), create_data((ts, 3))),
        # all matrix parameters will be estimated using em algorithm.
        (
            dict(params_3_3_static, state_dim=3, estimate_matrices="all"),
            create_data((ts, 3)),
        ),
        # test case 3 -
        #   state_dim = 2, measurement_dim = 3, params - params_2_3_,
        #   b is set with random ndarray of shape (2,).
        (
            dict(params_2_3_, state_dim=2, transition_offsets=create_data(2)),
            create_data((ts, 3)),
        ),
        # b, d and R will be estimated using em algorithm.
        (
            dict(
                params_2_3_,
                state_dim=2,
                transition_offsets=create_data(2),
                estimate_matrices=[
                    "measurement_offsets",
                    "transition_offsets",
                    "measurement_noise",
                ],
            ),
            create_data((ts, 3)),
        ),
        # test 4 -
        #   state_dim = 3, measurement_dim = 1, params are None
        (dict(state_dim=3), create_data((ts, 1), missing_values=True)),
        # F and Q will be estimated using em algorithm.
        (
            dict(state_dim=3, estimate_matrices=["state_transition", "process_noise"]),
            create_data((ts, 1), missing_values=True),
        ),
        # test 5 -
        #   state_dim = 1, measurement_dim = 1, params - params_1_1_arrays,
        #   b and d each set with random ndarray of shape (1,).
        (
            dict(
                params_1_1_arrays,
                state_dim=1,
                transition_offsets=create_data(1),
                measurement_offsets=create_data(1),
            ),
            create_data((ts, 1)),
        ),
        # P0 will be estimated using em algorithm.
        (
            dict(
                params_1_1_arrays,
                state_dim=1,
                estimate_matrices=["initial_state_covariance"],
                transition_offsets=create_data(1),
                measurement_offsets=create_data(1),
            ),
            create_data((ts, 1)),
        ),
        # test case 6 -
        #   state_dim = 1, measurement_dim = 1, params - params_1_1_lists
        #   b and d each set with a list of random ndarrays.
        #   transition_offsets (aka b) - each element of b is of shape
        #   (1,), b.length == 10.
        #   measurement_offsets (aka d) - each element of d is of shape
        #   (1,), d.length == 10.
        (
            dict(
                params_1_1_lists,
                state_dim=1,
                transition_offsets=rand_list(1),
                measurement_offsets=rand_list(1),
            ),
            create_data((ts, 1), missing_values=True),
        ),
        # test case 7 -
        #   state_dim = 3, measurement_dim = 1, params - params_3_1_lists
        #   b and d each set with a list of random ndarrays.
        #   transition_offsets (aka b) - each element of b is of shape
        #   (3,), b.length == 10.
        #   measurement_offsets (aka d) - each element of d is of shape
        #   (1,), d.length == 10.
        (
            dict(
                params_3_1_lists,
                state_dim=3,
                transition_offsets=rand_list(3),
                measurement_offsets=rand_list(1),
            ),
            create_data((ts, 1), missing_values=True),
        ),
    ],
)
def test_transform_and_smooth_pk(params, measurements):
    """Test KalmanFilterTransformerPK `fit` and `transform`.

    Creating two instances of KalmanFilterTransformerPK, one instance
    with parameter `denoising` set to False, and the other's set to True.
    Compare result with `pykalman`'s `filter` and `smooth`.
    """
    mask_measurements = np.ma.masked_invalid(np.copy(measurements))

    # adapter transformer
    adapter_transformer = KalmanFilterTransformerPK(
        **params
    )  # init_pykalman_adapter(state_dim=state_dim, **params)
    xt_adapter_transformer = adapter_transformer.fit_transform(measurements)

    # adapter denoising
    adapter_denoising = KalmanFilterTransformerPK(**params, denoising=True)
    xt_adapter_denoising = adapter_denoising.fit_transform(measurements)

    # pykalman
    kf_pykalman = init_kf_pykalman(measurement_dim=measurements.shape[1], **params)
    if "estimate_matrices" in params.keys():
        kf_pykalman = kf_pykalman.em(mask_measurements)
    xt_pykalman_transformer, _ = kf_pykalman.filter(mask_measurements)
    xt_pykalman_denoising, _ = kf_pykalman.smooth(mask_measurements)

    # test filter()
    assert np.array_equal(xt_adapter_transformer, xt_pykalman_transformer)

    # test smooth()
    assert np.array_equal(xt_adapter_denoising, xt_pykalman_denoising)


@pytest.mark.parametrize(
    "classes, params, measurements",
    [  # test case 1 -
        #   state_dim = 3, measurement_dim = 3, params - params_3_3_dynamic,
        #   X0 and H will be estimated using em algorithm.
        (
            [KalmanFilterTransformerPK, KalmanFilterTransformerFP],
            dict(
                params_3_3_dynamic,
                state_dim=3,
                estimate_matrices=["measurement_function", "initial_state"],
            ),
            create_data((ts, 3)),
        ),
        # test case 2 -
        #   state_dim = 3, measurement_dim = 3, params - params_3_3_static,
        #   all matrix parameters will be estimated using em algorithm.
        # test pykalman, transition_offsets (aka b) - is set with
        # random ndarray of shape (3,).
        (
            [KalmanFilterTransformerPK],
            dict(
                params_3_3_static,
                state_dim=3,
                transition_offsets=create_data(3),
                estimate_matrices=[
                    "state_transition",
                    "process_noise",
                    "measurement_offsets",
                    "transition_offsets",
                    "measurement_noise",
                    "measurement_function",
                    "initial_state",
                    "initial_state_covariance",
                ],
            ),
            create_data((ts, 3), missing_values=True),
        ),
        # test FilterPy
        (
            [KalmanFilterTransformerFP],
            dict(
                params_3_3_static,
                state_dim=3,
                estimate_matrices=[
                    "state_transition",
                    "process_noise",
                    "measurement_noise",
                    "measurement_function",
                    "initial_state",
                    "initial_state_covariance",
                ],
            ),
            create_data((ts, 3), missing_values=True),
        ),
        # test case 3 -
        #   state_dim = 2, measurement_dim = 3, params - params_2_3_
        # test both adapters, matrices Q, R, X0, P0 are estimated using em algorithm.
        (
            [KalmanFilterTransformerPK, KalmanFilterTransformerFP],
            dict(
                params_2_3_,
                state_dim=2,
                estimate_matrices=[
                    "process_noise",
                    "measurement_noise",
                    "initial_state",
                    "initial_state_covariance",
                ],
            ),
            create_data((ts, 3)),
        ),
        # test pykalman, transition_offsets (aka b) - is set with
        # random ndarray of shape (2,).
        (
            [KalmanFilterTransformerPK],
            dict(
                params_2_3_,
                state_dim=2,
                transition_offsets=create_data(2),
                estimate_matrices=["measurement_offsets", "transition_offsets"],
            ),
            create_data((ts, 3)),
        ),
        # test case 4 -
        #   state_dim = 3, measurement_dim = 1, params are None,
        #   all matrix parameters will be estimated using em algorithm.
        #   transition_offsets (aka b) - set with random ndarray of shape (3,).
        #   measurement_offsets (aka d) - set with random ndarray of shape (1,).
        # test pykalman
        (
            [KalmanFilterTransformerPK],
            dict(
                state_dim=3,
                transition_offsets=create_data(3),
                measurement_offsets=create_data(1),
                estimate_matrices="all",
            ),
            create_data((ts, 1), missing_values=True),
        ),
        # test case 5 -
        #   state_dim = 2, measurement_dim = 4, params are None,
        #   H and Q will be estimated using em algorithm.
        (
            [KalmanFilterTransformerPK, KalmanFilterTransformerFP],
            dict(
                state_dim=2, estimate_matrices=["measurement_function", "process_noise"]
            ),
            create_data((ts, 4), missing_values=True),
        ),
        # test case 6 -
        #   state_dim = 1, measurement_dim = 1, params - params_1_1_arrays
        # test pykalman, b and d each set with random ndarray of shape (1,).
        # b, d, X0, P0 will be estimated using em algorithm.
        (
            [KalmanFilterTransformerPK],
            dict(
                params_1_1_arrays,
                state_dim=1,
                transition_offsets=create_data(1),
                measurement_offsets=create_data(1),
                estimate_matrices=[
                    "measurement_offsets",
                    "transition_offsets",
                    "initial_state",
                    "initial_state_covariance",
                ],
            ),
            create_data((ts, 1)),
        ),
        # test pykalman, b and d each set with random ndarray of shape (10, 1).
        # R and H will be estimated using em algorithm.
        (
            [KalmanFilterTransformerPK],
            dict(
                params_1_1_arrays,
                state_dim=1,
                transition_offsets=create_data((ts, 1)),
                measurement_offsets=create_data((ts, 1)),
                estimate_matrices=["measurement_noise", "measurement_function"],
            ),
            create_data((ts, 1)),
        ),
        # test FilterPy, R, H, X0 and P0 will be estimated using em algorithm.
        (
            [KalmanFilterTransformerFP],
            dict(
                params_1_1_arrays,
                state_dim=1,
                estimate_matrices=[
                    "measurement_noise",
                    "measurement_function",
                    "initial_state",
                    "initial_state_covariance",
                ],
            ),
            create_data((ts, 1)),
        ),
        # test case 7 -
        #   state_dim = 1, measurement_dim = 1, params - params_1_1_lists,
        #   F will be estimated using em algorithm.
        (
            [KalmanFilterTransformerPK, KalmanFilterTransformerFP],
            dict(params_1_1_lists, state_dim=1, estimate_matrices=["state_transition"]),
            create_data((ts, 1), missing_values=True),
        ),
        # test case 8 -
        #   state_dim = 3, measurement_dim = 1, params - params_3_1_lists,
        #   X0 and P0 will be estimated using em algorithm.
        # test pykalman, b and d each set with a list of random ndarrays.
        # transition_offsets (aka b) - each element of b is of shape (3,),
        # b.length == 10.
        # measurement_offsets (aka d) - each element of d is of shape (1,),
        # d.length == 10.
        (
            [KalmanFilterTransformerPK],
            dict(
                params_3_1_lists,
                state_dim=3,
                transition_offsets=rand_list(3),
                measurement_offsets=rand_list(1),
                estimate_matrices=["initial_state", "initial_state_covariance"],
            ),
            create_data((ts, 1)),
        ),
        # test FilterPy
        (
            [KalmanFilterTransformerFP],
            dict(
                params_3_1_lists,
                state_dim=3,
                estimate_matrices=["initial_state", "initial_state_covariance"],
            ),
            create_data((ts, 1)),
        ),
    ],
)
def test_em(classes, params, measurements):
    """Test adapters matrix estimation.

    Call `fit` of input adapter/s, and compare all matrix parameters
    with `pykalman`'s matrix parameters returned from `em`.
    This test is useful for both KalmanFilterTransformerPK and
    KalmanFilterTransformerFP.
    """
    mask_measurements = np.ma.masked_invalid(np.copy(measurements))

    for _class in classes:
        adapter = _class(**params)
        adapter = adapter.fit(measurements)

        kf_pykalman = init_kf_pykalman(measurement_dim=measurements.shape[1], **params)
        kf_pykalman = kf_pykalman.em(X=mask_measurements)

        assert np.array_equal(adapter.F_, kf_pykalman.transition_matrices)
        assert np.array_equal(adapter.H_, kf_pykalman.observation_matrices)
        assert np.array_equal(adapter.Q_, kf_pykalman.transition_covariance)
        assert np.array_equal(adapter.R_, kf_pykalman.observation_covariance)
        assert np.array_equal(adapter.X0_, kf_pykalman.initial_state_mean)
        assert np.array_equal(adapter.P0_, kf_pykalman.initial_state_covariance)

        if hasattr(adapter, "transition_offsets_"):
            assert np.array_equal(
                adapter.transition_offsets_, kf_pykalman.transition_offsets
            )
        if hasattr(adapter, "measurement_offsets_"):
            assert np.array_equal(
                adapter.measurement_offsets_, kf_pykalman.observation_offsets
            )


@pytest.mark.parametrize(
    "classes, params, measurements",
    [  # test case 1 -
        #   state_dim = 3, measurement_dim = 3, params - params_3_3_dynamic
        # bad input :
        # typo in element of `estimate_matrices` : sstate_transition instead of
        # state_transition
        (
            [KalmanFilterTransformerPK, KalmanFilterTransformerFP],
            dict(
                params_3_3_dynamic,
                state_dim=3,
                estimate_matrices=["sstate_transition", "process_noise"],
            ),
            create_data((ts, 3)),
        ),
        # test case 2 -
        #   state_dim = 3, measurement_dim = 3, params - params_3_3_dynamic
        # bad input:
        # wrong b shape: set to ndarray of shape (10, 2) when should be (10, 3) or (3,)
        (
            [KalmanFilterTransformerPK],
            dict(
                params_3_3_dynamic,
                state_dim=3,
                estimate_matrices=[
                    "process_noise",
                    "measurement_noise",
                    "initial_state",
                    "initial_state_covariance",
                ],
                transition_offsets=create_data((ts, 2)),
            ),
            create_data((ts, 3), missing_values=True),
        ),
        # test case 3 -
        #   state_dim = 2, measurement_dim = 3, params - params_2_3_
        # bad input:
        # wrong d shape: set to ndarray of shape (11, 2) when should be (10, 2) or (2,)
        (
            [KalmanFilterTransformerPK],
            dict(params_2_3_, state_dim=2, measurement_offsets=create_data((11, 2))),
            create_data((ts, 3)),
        ),
        # test case 4 -
        #   state_dim = 2, measurement_dim = 3, params - params_2_3_
        # bad input:
        # typo in element of `estimate_matrices` : transition_offset instead of
        # transition_offsets
        (
            [KalmanFilterTransformerPK, KalmanFilterTransformerFP],
            dict(params_2_3_, state_dim=2, estimate_matrices=["transition_offset"]),
            create_data((ts, 3)),
        ),
        # test case 5 -
        #   state_dim = 3, measurement_dim = 5, params - are None
        #   except state_transition.
        # bad input:
        # wrong F shape: set to ndarray of shape (10, 3, 1) when should be
        # (10, 3, 3) or (3, 3)
        (
            [KalmanFilterTransformerPK, KalmanFilterTransformerFP],
            dict(state_dim=3, state_transition=create_data((ts, 3, 1))),
            create_data((ts, 5), missing_values=True),
        ),
        # test case 6 -
        #   state_dim = 4, measurement_dim = 4, params - are None
        # bad input:
        # typo in element of `estimate_matrices` : measurement_functions instead
        # of measurement_function
        (
            [KalmanFilterTransformerPK, KalmanFilterTransformerFP],
            dict(state_dim=4, estimate_matrices=["measurement_functions"]),
            create_data((ts, 4), missing_values=True),
        ),
        # bad input:
        # KalmanFilterTransformerFP does not estimate matrix transition_offsets.
        (
            [KalmanFilterTransformerFP],
            dict(state_dim=4, estimate_matrices=["transition_offsets"]),
            create_data((ts, 4), missing_values=True),
        ),
        # bad input:
        # KalmanFilterTransformerFP does not estimate matrix measurement_offsets.
        (
            [KalmanFilterTransformerFP],
            dict(state_dim=4, estimate_matrices=["measurement_offsets"]),
            create_data((ts, 4), missing_values=True),
        ),
        # bad input:
        # KalmanFilterTransformerFP does not estimate matrix control_transition.
        (
            [KalmanFilterTransformerFP],
            dict(state_dim=4, estimate_matrices=["control_transition"]),
            create_data((ts, 4), missing_values=True),
        ),
        # test case 7 -
        #   state_dim = 1, measurement_dim = 1, params - params_1_1_arrays
        # bad input:
        # typo in element of `estimate_matrices` : initial_state_mean
        # instead of initial_state
        (
            [KalmanFilterTransformerPK, KalmanFilterTransformerFP],
            dict(
                params_1_1_arrays,
                state_dim=1,
                estimate_matrices=[
                    "state_transition",
                    "initial_state_mean",
                    "initial_state_covariance",
                ],
            ),
            create_data((ts, 1)),
        ),
        # test case 8 -
        #   state_dim = 2, measurement_dim = 3, params - as described
        # bad inputs:
        # typo in element of `estimate_matrices` : covariance instead of
        # initial_state_covariance
        # wrong H shape: set to ndarray of shape (2, 3) when should be
        # (10, 3, 2) or (3, 2)
        (
            [KalmanFilterTransformerPK, KalmanFilterTransformerFP],
            dict(
                state_dim=2,
                estimate_matrices=["measurement_noise", "covariance"],
                state_transition=create_data((3, 3)),
                measurement_function=create_data((2, 3)),
            ),
            create_data((ts, 3)),
        ),
    ],
)
def test_bad_inputs(classes, params, measurements):
    """Test adapters bad inputs error handling.

    Call `fit` of input adapter/s, and pass if ValueError
    was thrown.
    This test is useful for both KalmanFilterTransformerPK
    and KalmanFilterTransformerFP.
    """
    with pytest.raises(ValueError):
        for _class in classes:
            adapter = _class(**params)
            adapter.fit(X=measurements)


@pytest.mark.parametrize(
    "params, measurements, y",
    [  # test case 1 -
        #   state_dim = 3, measurement_dim = 3, params - params_3_3_dynamic
        (dict(params_3_3_dynamic, state_dim=3), create_data((ts, 3)), None),
        # control_transition (aka G or B) is set with random ndarray of shape (3, 3).
        # y is set with random ndarray of shape (10, 3).
        # H and X0 will be estimated using em algorithm.
        (
            dict(
                params_3_3_dynamic,
                state_dim=3,
                control_transition=create_data((3, 3)),
                estimate_matrices=["measurement_function", "initial_state"],
            ),
            create_data((ts, 3)),
            create_data((ts, 3)),
        ),
        # test case 2 -
        #   state_dim = 3, measurement_dim = 3, params - params_3_3_static
        # control_transition (aka G or B) is set with a list of 10 random
        # ndarrays, each of shape (3, 2).
        # y is set with random ndarray of shape (2,).
        (
            dict(params_3_3_static, state_dim=3, control_transition=rand_list((3, 2))),
            create_data((ts, 3)),
            create_data(2),
        ),
        # all matrix parameters will be estimated using em algorithm.
        # control_transition (aka G or B) is set with random ndarray of shape (3, 1).
        # y is set with random ndarray of shape (1,).
        (
            dict(
                params_3_3_static,
                state_dim=3,
                estimate_matrices="all",
                control_transition=create_data((3, 1)),
            ),
            create_data((ts, 3)),
            create_data(1),
        ),
        # test case 3 -
        #   state_dim = 2, measurement_dim = 3, params - params_2_3_,
        # y is set with random ndarray of shape (10, 4).
        (dict(params_2_3_, state_dim=2), create_data((ts, 3)), create_data((ts, 4))),
        # R will be estimated using em algorithm.
        (
            dict(params_2_3_, state_dim=2, estimate_matrices=["measurement_noise"]),
            create_data((ts, 3)),
            None,
        ),
        # test case 4 -
        #   state_dim = 3, measurement_dim = 1, params are None
        # control_transition (aka G or B) - is set with random ndarray of shape (3, 2).
        # should raise a warning and control_transition will be
        # ignored during calculation.
        (
            dict(state_dim=3, control_transition=create_data((3, 2))),
            create_data((ts, 1), missing_values=True),
            None,
        ),
        # F and Q will be estimated using em algorithm.
        (
            dict(state_dim=3, estimate_matrices=["state_transition", "process_noise"]),
            create_data((ts, 1), missing_values=True),
            None,
        ),
        # test case 5 -
        #   state_dim = 1, measurement_dim = 1, params - params_1_1_arrays,
        # control_transition is set with random ndarray of shape (10, 1).
        # y is set with random ndarray of shape (10, 1).
        (
            dict(
                params_1_1_arrays,
                state_dim=1,
                control_transition=create_data((ts, 1, 1)),
            ),
            create_data((ts, 1)),
            create_data((ts, 1)),
        ),
        # P0 will be estimated using em algorithm.
        # control_transition is set with a list of 10 random ndarrays of shape (1,).
        # y is set with random ndarray of shape (10, 1).
        (
            dict(
                params_1_1_arrays,
                state_dim=1,
                estimate_matrices=["initial_state_covariance"],
                control_transition=rand_list((1, 1)),
            ),
            create_data((ts, 1)),
            create_data((ts, 1)),
        ),
        # test case 6 -
        #   state_dim = 1, measurement_dim = 1, params - params_1_1_lists
        # y is set with random ndarray of shape (2,).
        (
            dict(params_1_1_lists, state_dim=1),
            create_data((ts, 1), missing_values=True),
            create_data(2),
        ),
        # test case 7 -
        #   state_dim = 3, measurement_dim = 1, params - params_3_1_lists
        # control_transition is set with random ndarray of shape (3, 3).
        # y is set with random ndarray of shape (10, 1).
        (
            dict(params_3_1_lists, state_dim=3, control_transition=create_data((3, 3))),
            create_data((ts, 1), missing_values=True),
            create_data((ts, 3)),
        ),
    ],
)
def test_transform_and_smooth_fp(params, measurements, y):
    """Test KalmanFilterTransformerFP `fit` and `transform`.

    Creating two instances of KalmanFilterTransformerFP, one instance
    with parameter `denoising` set to False, and the other's set to True.
    Compare result with `FilterPy`'s `batch_filter` and `rts_smoother`.
    """
    from filterpy.kalman.kalman_filter import batch_filter, rts_smoother

    # initiate KalmanFilterTransformerFP with denoising=False
    # fit and transform
    adapter_transformer = KalmanFilterTransformerFP(**params)
    adapter_transformer = adapter_transformer.fit(measurements, y=y)
    xt_transformer_adapter = adapter_transformer.transform(measurements, y=y)

    # initiating KalmanFilterTransformerFP with denoising=True,
    # fit and transform
    adapter_smoother = KalmanFilterTransformerFP(denoising=True, **params)
    adapter_smoother = adapter_smoother.fit(measurements, y=y)
    xt_smoother_adapter = adapter_smoother.transform(measurements, y=y)

    # get data in a form compatible to FilterPy. call batch_filter
    matrices, fp_measurements = init_kf_filterpy(
        measurements=measurements, adapter=adapter_transformer, y=y
    )
    (means, covs, _, _) = batch_filter(**matrices, zs=fp_measurements)

    # test transformer
    assert_array_almost_equal(xt_transformer_adapter, means)

    # test smoother
    xt_smoother_filterpy = rts_smoother(
        Xs=means, Ps=covs, Fs=matrices["Fs"], Qs=matrices["Qs"]
    )[0]
    assert_array_almost_equal(xt_smoother_adapter, xt_smoother_filterpy)



# -*- coding: utf-8 -*-
import matplotlib.colors as colorplt
import matplotlib.pyplot as plt
import numpy as np

from sktime.distances._distance import distance_alignment_path, pairwise_distance

gray_cmap = colorplt.LinearSegmentedColormap.from_list("", ["#c9cacb", "white"])


def _path_mask(cost_matrix, path, ax, theme=gray_cmap):
    plot_matrix = np.zeros_like(cost_matrix)
    max_size = max(cost_matrix.shape)
    for i in range(max_size):
        for j in range(max_size):
            if (i, j) in path:
                plot_matrix[i, j] = 1.0
            elif cost_matrix[i, j] == np.inf:
                plot_matrix[i, j] = 0.0
            else:
                plot_matrix[i, j] = 0.25

    for i in range(max_size):
        for j in range(max_size):
            c = cost_matrix[j, i]
            ax.text(i, j, str(round(c, 2)), va="center", ha="center", size=10)
            ax.text(i, j, str(round(c, 2)), va="center", ha="center", size=10)

    ax.matshow(plot_matrix, cmap=theme)


def _pairwise_path(x, y, metric):
    pw_matrix = pairwise_distance(x, y, metric=metric)
    path = []
    for i in range(pw_matrix.shape[0]):
        for j in range(pw_matrix.shape[1]):
            if i == j:
                path.append((i, j))
    return path, pw_matrix.trace(), pw_matrix


def _plot_path(
    x: np.ndarray,
    y: np.ndarray,
    metric: str,
    dist_kwargs: dict = None,
    title: str = "",
    plot_over_pw: bool = False,
):
    if dist_kwargs is None:
        dist_kwargs = {}
    try:
        path, dist, cost_matrix = distance_alignment_path(
            x, y, metric=metric, return_cost_matrix=True, **dist_kwargs
        )

        if metric == "lcss":
            _path = []
            for tup in path:
                _path.append(tuple(x + 1 for x in tup))
            path = _path

        if plot_over_pw is True:
            if metric == "lcss":
                pw = pairwise_distance(x, y, metric="euclidean")
                cost_matrix = np.zeros_like(cost_matrix)
                cost_matrix[1:, 1:] = pw
            else:
                pw = pairwise_distance(x, y, metric="squared")
                cost_matrix = pw
    except NotImplementedError:
        path, dist, cost_matrix = _pairwise_path(x, y, metric)

    plt.figure(1, figsize=(8, 8))
    x_size = x.shape[0]

    # definitions for the axes
    left, bottom = 0.01, 0.1
    w_ts = h_ts = 0.2
    left_h = left + w_ts + 0.02
    width = height = 0.65
    bottom_h = bottom + height + 0.02

    rect_s_y = [left, bottom, w_ts, height]
    rect_gram = [left_h, bottom, width, height]
    rect_s_x = [left_h, bottom_h, width, h_ts]

    ax_gram = plt.axes(rect_gram)
    ax_s_x = plt.axes(rect_s_x)
    ax_s_y = plt.axes(rect_s_y)

    _path_mask(cost_matrix, path, ax_gram)
    ax_gram.axis("off")
    ax_gram.autoscale(False)
    # ax_gram.plot([j for (i, j) in path], [i for (i, j) in path], "w-",
    #              linewidth=3.)

    ax_s_x.plot(np.arange(x_size), y, "b-", linewidth=3.0, color="#818587")
    ax_s_x.axis("off")
    ax_s_x.set_xlim((0, x_size - 1))

    ax_s_y.plot(-x, np.arange(x_size), "b-", linewidth=3.0, color="#818587")
    ax_s_y.axis("off")
    ax_s_y.set_ylim((0, x_size - 1))

    ax_s_x.set_title(title, size=10)

    return plt


def _plot_alignment(x, y, metric, dist_kwargs: dict = None, title: str = ""):
    if dist_kwargs is None:
        dist_kwargs = {}
    try:
        path, dist, cost_matrix = distance_alignment_path(
            x, y, metric=metric, return_cost_matrix=True, **dist_kwargs
        )
    except NotImplementedError:
        path, dist, cost_matrix = _pairwise_path(x, y, metric)

    plt.figure(1, figsize=(8, 8))

    plt.plot(x, "b-", color="black")
    plt.plot(y, "g-", color="black")

    for positions in path:
        try:
            plt.plot(
                [positions[0], positions[1]],
                [x[positions[0]], y[positions[1]]],
                "--",
                color="#818587",
            )
        except:
            continue
    plt.legend()
    plt.title(title)

    plt.tight_layout()
    return plt


if __name__ == "__main__":
    x = np.array(
        [
            -0.7553383207,
            0.4460987596,
            1.197682907,
            0.1714334808,
            0.5639929213,
            0.6891222874,
            1.793828873,
            0.06570866314,
            0.2877381702,
            1.633620422,
        ]
    )

    y = np.array(
        [
            0.01765193577,
            1.536784164,
            -0.1413292622,
            -0.7609346135,
            -0.1767363331,
            -2.192007072,
            -0.1933165696,
            -0.4648166839,
            -0.9444888843,
            -0.239523623,
        ]
    )
    import os

    def _save_plt(plt):
        plt[0].savefig(f"{metric_path}/{plt[1]}")
        plt[0].cla()
        plt[0].clf()

    if not os.path.exists("./plots"):
        os.makedirs("./plots")
    metrics = [
        "euclidean",
        "erp",
        "edr",
        "lcss",
        "squared",
        "dtw",
        "ddtw",
        "wdtw",
        "wddtw",
        "msm",
        "twe",
    ]
    # metrics = ['lcss']
    for metric in metrics:
        metric_path = f"./plots/{metric}"
        if not os.path.exists(metric_path):
            os.makedirs(metric_path)

        _save_plt(
            (
                _plot_path(x, y, metric, {"epsilon": 1.0}),
                f"{metric}_path_through_cost_matrix",
            )
        )
        _save_plt(
            (
                _plot_path(x, y, metric, {"window": 0.2, "epsilon": 1.0}),
                f"{metric}_path_through_20_cost_matrix",
            )
        )

        if metric == "wdtw":
            g_val = [0.2, 0.3]
            for g in g_val:
                file_save = str(g).split(".")
                _save_plt(
                    (
                        _plot_path(x, y, metric, {"g": g}),
                        f"{metric}_path_through_g{file_save[1]}_cost_matrix",
                    )
                )

        _save_plt((_plot_alignment(x, y, metric), f"{metric}_alignment"))
        _save_plt(
            (_plot_alignment(x, y, metric, {"window": 0.2}), f"{metric}_alignment_20")
        )



# -*- coding: utf-8 -*-
"""Proximity Forest time series classifier.

A decision tree forest which uses distance measures to partition data.
B. Lucas and A. Shifaz, C. Pelletier, L. O’Neill, N. Zaidi, B. Goethals,
F. Petitjean and G. Webb
Proximity Forest: an effective and scalable distance-based classifier for
time series,
Data Mining and Knowledge Discovery, 33(3): 607-635, 2019
"""

# linkedin.com/goastler; github.com/goastler
# github.com/moradisten
__author__ = ["gastler", "Morad A. Azaz"]
__all__ = ["ProximityForest", "_CachedTransformer", "ProximityStump", "ProximityTree"]

from logging import exception

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from scipy import stats
from sklearn.preprocessing import LabelEncoder, normalize
from sklearn.utils import check_random_state

from sktime.classification.base import BaseClassifier
from sktime.datatypes._panel._convert import (
    from_nested_to_2d_array,
    from_nested_to_3d_numpy,
)
from sktime.distances import (  # twe_distance,
    dtw_distance,
    erp_distance,
    lcss_distance,
    msm_distance,
    wdtw_distance,
)
from sktime.transformations.base import _PanelToPanelTransformer
from sktime.transformations.panel.summarize import DerivativeSlopeTransformer
from sktime.utils.validation.panel import check_X, check_X_y

# todo unit tests / sort out current unit tests
# todo logging package rather than print to screen
# todo get params avoid func pointer - use name
# todo set params use func name or func pointer
# todo constructor accept str name func / pointer
# todo duck-type functions


def stdp(X):
    """Proximity forest util function (deprecated)."""
    warn(
        "This function has moved to classification/distance_based/_proximity_forest as "
        "a private function. This version will be removed in V0.10",
        FutureWarning,
    )
    sum = 0
    sum_sq = 0
    num_instances = X.shape[0]
    num_dimensions = X.shape[1]
    num_values = 0
    for instance_index in range(0, num_instances):
        for dimension_index in range(0, num_dimensions):
            instance = X.iloc[instance_index, dimension_index]
            for value in instance:
                num_values += 1
                sum += value
                sum_sq += value**2  # todo missing values NaN messes
                # this up!
    mean = sum / num_values
    stdp = np.math.sqrt(sum_sq / num_values - mean**2)
    return stdp


# find index of min value in array, randomly breaking ties
def _arg_min(array, rand, getter=None):
    """Proximity forest util function."""
    return rand.choice(arg_mins(array, getter))


def max_instance_length(X):
    """Proximity forest util function."""
    max_length = len(X.iloc[0, 0])
    # max = -1
    # for dimension in range(0, instances.shape[1]):
    #     length = max_instance_dimension_length(instances, dimension)
    #     if length > max:
    #         max = length
    return max_length


class _CachedTransformer(_PanelToPanelTransformer):
    """Transformer that transforms data and adds the transformed version to a cache.

    If the transformation is called again on already seen data the data is
    fetched from the cache rather than performing the expensive transformation.

    Parameters
    ----------
    transformer : the transformer to transform uncached data
    Attributes
    ----------
    cache       : location to store transforms seen before for fast look up

    """

    _required_parameters = ["transformer"]

    def __init__(self, transformer):
        self.cache = {}
        self.transformer = transformer
        super(_CachedTransformer, self).__init__()

    def clear(self):
        """Clear the cache."""
        self.cache = {}

    def transform(self, X, y=None):
        """
        Fit transformer, creating a cache for transformation.

        Parameters
        ----------
        X : pandas DataFrame of shape [n_instances, n_features]
            Input data
        y : pandas Series, shape (n_instances), optional
            Targets for supervised learning.

        Returns
        -------
        cached_instances.
        """
        # for each instance, get transformed instance from cache or
        # transform and add to cache
        cached_instances = {}
        uncached_indices = []
        for index in X.index.values:
            try:
                cached_instances[index] = self.cache[index]
            except Exception:
                uncached_indices.append(index)
        if len(uncached_indices) > 0:
            uncached_instances = X.loc[uncached_indices, :]
            transformed_uncached_instances = self.transformer.fit_transform(
                uncached_instances
            )
            transformed_uncached_instances.index = uncached_instances.index
            transformed_uncached_instances = transformed_uncached_instances.to_dict(
                "index"
            )
            self.cache.update(transformed_uncached_instances)
            cached_instances.update(transformed_uncached_instances)
        cached_instances = pd.DataFrame.from_dict(cached_instances, orient="index")
        return cached_instances

    def __str__(self):
        """Transform string."""
        return self.transformer.__str__()


def _derivative_distance(distance_measure, transformer):
    """Take derivative before conducting distance measure.

    :param distance_measure: the distance measure to use
    :param transformer: the transformer to use
    :return: a distance measure function with built in transformation
    """

    def distance(instance_a, instance_b, **params):
        df = pd.DataFrame([instance_a, instance_b])
        df = transformer.transform(X=df)
        instance_a = df.iloc[0, :]
        instance_b = df.iloc[1, :]
        return distance_measure(instance_a, instance_b, **params)

    return distance


def distance_predefined_params(distance_measure, **params):
    """Conduct distance measurement with a predefined set of parameters.

    :param distance_measure: the distance measure to use
    :param params: the parameters to use in the distance measure
    :return: a distance measure with no parameters
    """

    def distance(instance_a, instance_b):
        return distance_measure(instance_a, instance_b, **params)

    return distance


def numba_wrapper(distance_measure):
    """Wrap a distance measure in cython conversion.

    (to 1 column per dimension format)
    :param distance_measure: distance measure to wrap
    :return: a distance measure which automatically formats data for cython
    distance measures
    """

    def distance(instance_a, instance_b, **params):
        # find distance
        instance_a = from_nested_to_2d_array(
            instance_a, return_numpy=True
        )  # todo use specific
        # dimension rather than whole
        # thing?
        instance_b = from_nested_to_2d_array(
            instance_b, return_numpy=True
        )  # todo use specific
        # dimension rather than whole thing?
        instance_a = np.transpose(instance_a)
        instance_b = np.transpose(instance_b)
        return distance_measure(instance_a, instance_b, **params)

    return distance


def pure(y):
    """Test whether a set of class labels are pure (i.e. all the same).

    Parameters
    ----------
    y : 1d array like
        array of class labels

    Returns
    -------
    result : boolean
        whether the set of class labels is pure
    """
    # get unique class labels
    unique_class_labels = np.unique(np.array(y))
    # if more than 1 unique then not pure
    return len(unique_class_labels) <= 1


def gini_gain(y, y_subs):
    """Get gini score of a split, i.e. the gain from parent to children.

    Parameters
    ----------
    y : 1d array like
        array of class labels at parent
    y_subs : list of 1d array like
        list of array of class labels, one array per child

    Returns
    -------
    score : float
        gini score of the split from parent class labels to children. Note a
        higher score means better gain,
        i.e. a better split
    """
    y = np.array(y)
    # find number of instances overall
    parent_n_instances = y.shape[0]
    # if parent has no instances then is pure
    if parent_n_instances == 0:
        for child in y_subs:
            if len(child) > 0:
                raise ValueError("children populated but parent empty")
        return 0.5
    # find gini for parent node
    score = gini(y)
    # sum the children's gini scores
    for index in range(len(y_subs)):
        child_class_labels = y_subs[index]
        # ignore empty children
        if len(child_class_labels) > 0:
            # find gini score for this child
            child_score = gini(child_class_labels)
            # weight score by proportion of instances at child compared to
            # parent
            child_size = len(child_class_labels)
            child_score *= child_size / parent_n_instances
            # add to cumulative sum
            score -= child_score
    return score


def gini(y):
    """Get gini score at a specific node.

    Parameters
    ----------
    y : 1d numpy array
        array of class labels

    Returns
    -------
    score : float
        gini score for the set of class labels (i.e. how pure they are). A
        larger score means more impurity. Zero means
        pure.
    """
    y = np.array(y)
    # get number instances at node
    n_instances = y.shape[0]
    if n_instances > 0:
        # count each class
        unique_class_labels, class_counts = np.unique(y, return_counts=True)
        # subtract class entropy from current score for each class
        class_counts = np.divide(class_counts, n_instances)
        class_counts = np.power(class_counts, 2)
        class_counts_sum = np.sum(class_counts)
        return 1 - class_counts_sum
    else:
        # y is empty, therefore considered pure
        raise ValueError(" y empty")


def get_one_exemplar_per_class_proximity(proximity):
    """Unpack proximity object into X, y and random_state for picking exemplars.

    Parameters
    ----------
    proximity : Proximity object
        Proximity like object containing the X, y and random_state variables
        required for picking exemplars.

    Returns
    -------
    result : function
        function choosing one exemplar per class
    """
    return get_one_exemplar_per_class(proximity.X, proximity.y, proximity.random_state)


def get_one_exemplar_per_class(X, y, random_state):
    """Pick one exemplar instance per class in the dataset.

    Parameters
    ----------
    X : array-like or sparse matrix of shape = [n_instances, n_columns]
            The training input samples.  If a Pandas data frame is passed,
            the column _dim_to_use is extracted
    y : array-like, shape = [n_samples] or [n_samples, n_outputs]
        The class labels.
    random_state : numpy RandomState
        a random state for sampling random numbers

    Returns
    -------
    chosen_instances : list
        list of the chosen exemplar instances.
    chosen_class_labels : array
        list of corresponding class labels for each of the chosen exemplar
        instances.
    """
    # find unique class labels
    unique_class_labels = np.unique(y)
    n_unique_class_labels = len(unique_class_labels)
    chosen_instances = [None] * n_unique_class_labels
    # for each class randomly choose and instance
    for class_label_index in range(n_unique_class_labels):
        class_label = unique_class_labels[class_label_index]
        # filter class labels for desired class and get indices
        indices = np.argwhere(y == class_label)
        # flatten numpy output
        indices = np.ravel(indices)
        # random choice
        index = random_state.choice(indices)
        # record exemplar instance and class label
        instance = X.iloc[index, :]
        chosen_instances[class_label_index] = instance
    # convert lists to numpy arrays
    return chosen_instances, unique_class_labels


def dtw_distance_measure_getter(X):
    """Generate the dtw distance measure.

    :param X: dataset to derive parameter ranges from
    :return: distance measure and parameter range dictionary
    """
    return {
        "distance_measure": [numba_wrapper(dtw_distance)],
        "w": stats.uniform(0, 0.25),
    }


def msm_distance_measure_getter(X):
    """Generate the msm distance measure.

    :param X: dataset to derive parameter ranges from
    :return: distance measure and parameter range dictionary
    """
    n_dimensions = 1  # todo use other dimensions
    return {
        "distance_measure": [numba_wrapper(dtw_distance)],
        "dim_to_use": stats.randint(low=0, high=n_dimensions),
        "c": [
            0.01,
            0.01375,
            0.0175,
            0.02125,
            0.025,
            0.02875,
            0.0325,
            0.03625,
            0.04,
            0.04375,
            0.0475,
            0.05125,
            0.055,
            0.05875,
            0.0625,
            0.06625,
            0.07,
            0.07375,
            0.0775,
            0.08125,
            0.085,
            0.08875,
            0.0925,
            0.09625,
            0.1,
            0.136,
            0.172,
            0.208,
            0.244,
            0.28,
            0.316,
            0.352,
            0.388,
            0.424,
            0.46,
            0.496,
            0.532,
            0.568,
            0.604,
            0.64,
            0.676,
            0.712,
            0.748,
            0.784,
            0.82,
            0.856,
            0.892,
            0.928,
            0.964,
            1,
            1.36,
            1.72,
            2.08,
            2.44,
            2.8,
            3.16,
            3.52,
            3.88,
            4.24,
            4.6,
            4.96,
            5.32,
            5.68,
            6.04,
            6.4,
            6.76,
            7.12,
            7.48,
            7.84,
            8.2,
            8.56,
            8.92,
            9.28,
            9.64,
            10,
            13.6,
            17.2,
            20.8,
            24.4,
            28,
            31.6,
            35.2,
            38.8,
            42.4,
            46,
            49.6,
            53.2,
            56.8,
            60.4,
            64,
            67.6,
            71.2,
            74.8,
            78.4,
            82,
            85.6,
            89.2,
            92.8,
            96.4,
            100,
        ],
    }


def erp_distance_measure_getter(X):
    """Generate the erp distance measure.

    :param X: dataset to derive parameter ranges from
    :return: distance measure and parameter range dictionary
    """
    stdp = _stdp(X)
    instance_length = max_instance_length(X)  # todo should this use the max instance
    # length for unequal length dataset instances?
    max_raw_warping_window = np.floor((instance_length + 1) / 4)
    n_dimensions = 1  # todo use other dimensions
    return {
        "distance_measure": [numba_wrapper(erp_distance)],
        "dim_to_use": stats.randint(low=0, high=n_dimensions),
        "g": stats.uniform(0.2 * stdp, 0.8 * stdp - 0.2 * stdp),
        "band_size": stats.randint(low=0, high=max_raw_warping_window + 1)
        # scipy stats randint is exclusive on the max value, hence + 1
    }


def lcss_distance_measure_getter(X):
    """Generate the lcss distance measure.

    :param X: dataset to derive parameter ranges from
    :return: distance measure and parameter range dictionary
    """
    stdp = _stdp(X)
    instance_length = max_instance_length(X)  # todo should this use the max instance
    # length for unequal length dataset instances?
    max_raw_warping_window = np.floor((instance_length + 1) / 4)
    n_dimensions = 1  # todo use other dimensions
    return {
        "distance_measure": [numba_wrapper(lcss_distance)],
        "dim_to_use": stats.randint(low=0, high=n_dimensions),
        "epsilon": stats.uniform(0.2 * stdp, stdp - 0.2 * stdp),
        # scipy stats randint is exclusive on the max value, hence + 1
        "delta": stats.randint(low=0, high=max_raw_warping_window + 1),
    }


def twe_distance_measure_getter(X):
    """Generate the twe distance measure.

    :param X: dataset to derive parameter ranges from
    :return: distance measure and parameter range dictionary
    """
    return {
        "distance_measure": [numba_wrapper(twe_distance)],
        "penalty": [
            0,
            0.011111111,
            0.022222222,
            0.033333333,
            0.044444444,
            0.055555556,
            0.066666667,
            0.077777778,
            0.088888889,
            0.1,
        ],
        "stiffness": [0.00001, 0.0001, 0.0005, 0.001, 0.005, 0.01, 0.05, 0.1, 0.5, 1],
    }


def wdtw_distance_measure_getter(X):
    """Generate the wdtw distance measure.

    :param X: dataset to derive parameter ranges from
    :return: distance measure and parameter range dictionary
    """
    return {
        "distance_measure": [numba_wrapper(weighted_dtw_distance)],
        "g": stats.uniform(0, 1),
    }


def euclidean_distance_measure_getter(X):
    """Generate the ed distance measure.

    :param X: dataset to derive parameter ranges from
    :return: distance measure and parameter range dictionary
    """
    return {"distance_measure": [numba_wrapper(dtw_distance)], "w": [0]}


def setup_wddtw_distance_measure_getter(transformer):
    """Generate the wddtw distance measure by baking the derivative transformer.

    into the wdtw distance measure
    :param transformer: the transformer to use
    :return: a getter to produce the distance measure
    """

    def getter(X):
        return {
            "distance_measure": [
                _derivative_distance(numba_wrapper(weighted_dtw_distance), transformer)
            ],
            "g": stats.uniform(0, 1),
        }

    return getter


def setup_ddtw_distance_measure_getter(transformer):
    """Generate the ddtw distance measure by baking the derivative transformer.

    into the dtw distance measure
    :param transformer: the transformer to use
    :return: a getter to produce the distance measure
    """

    def getter(X):
        return {
            "distance_measure": [
                _derivative_distance(numba_wrapper(dtw_distance), transformer)
            ],
            "w": stats.uniform(0, 0.25),
        }

    return getter


def setup_all_distance_measure_getter(proximity):
    """Set all distance measure getter functions from a proximity object.

    :param proximity: a PT / PF / PS
    :return: a list of distance measure getters
    """
    transformer = _CachedTransformer(DerivativeSlopeTransformer())
    distance_measure_getters = [
        euclidean_distance_measure_getter,
        dtw_distance_measure_getter,
        setup_ddtw_distance_measure_getter(transformer),
        wdtw_distance_measure_getter,
        setup_wddtw_distance_measure_getter(transformer),
        msm_distance_measure_getter,
        lcss_distance_measure_getter,
        erp_distance_measure_getter,
        #        twe_distance_measure_getter,
    ]

    def pick_rand_distance_measure(proximity):
        """Generate a distance measure from a range of parameters.

        :param proximity: proximity object containing distance measures,
        ranges and dataset
        :return: a distance measure with no parameters
        """
        random_state = proximity.random_state
        X = proximity.X
        distance_measure_getter = random_state.choice(distance_measure_getters)
        distance_measure_perm = distance_measure_getter(X)
        param_perm = pick_rand_param_perm_from_dict(distance_measure_perm, random_state)
        distance_measure = param_perm["distance_measure"]
        del param_perm["distance_measure"]
        return distance_predefined_params(distance_measure, **param_perm)

    return pick_rand_distance_measure


def pick_rand_param_perm_from_dict(param_pool, random_state):
    """Pick a parameter permutation.

    Given a list of dictionaries contain potential values OR a list of values OR a
    distribution of values (a distribution must have the .rvs() function to
    sample values)

    param_pool : list of dicts OR list OR distribution
        parameters in the same format as GridSearchCV from scikit-learn.
        example:
        param_grid = [
          {'C': [1, 10, 100, 1000], 'kernel': ['linear']},
          {'C': [1, 10, 100, 1000], 'gamma': [{'C': [1, 10, 100, 1000],
          'kernel': ['linear']}],
          'kernel': ['rbf']},
         ]

    Returns
    -------
    param_perm : dict
        distance measure and corresponding parameters in dictionary format
    """
    # construct empty permutation
    param_perm = {}
    # for each parameter
    for param_name, param_values in param_pool.items():
        # if it is a list
        if isinstance(param_values, list):
            # randomly pick a value
            param_value = param_values[random_state.randint(len(param_values))]
            # if the value is another dict then get a random parameter
            # permutation from that dict (recursive over
            # 2 funcs)
            # if isinstance(param_value, dict): # no longer require
            # recursive param perms
            #     param_value = _pick_param_permutation(param_value,
            #     random_state)
        # else if parameter is a distribution
        elif hasattr(param_values, "rvs"):
            # sample from the distribution
            param_value = param_values.rvs(random_state=random_state)
        else:
            # otherwise we don't know how to obtain a value from the parameter
            raise Exception("unknown type of parameter pool")
        # add parameter name and value to permutation
        param_perm[param_name] = param_value
    return param_perm


def pick_rand_param_perm_from_list(params, random_state):
    """Get a random parameter permutation.

     Providing a distance measure and corresponding parameters

    params : list of dicts
        parameters in the same format as GridSearchCV from scikit-learn.
        example:
        param_grid = [
          {'C': [1, 10, 100, 1000], 'kernel': ['linear']},
          {'C': [1, 10, 100, 1000], 'gamma': [{'C': [1, 10, 100, 1000],
          'kernel': ['linear']}], 'kernel': ['rbf']},
         ]

    Returns
    -------
    permutation : dict
        distance measure and corresponding parameters in dictionary format
    """
    #
    param_pool = random_state.choice(params)
    permutation = pick_rand_param_perm_from_dict(param_pool, random_state)
    return permutation


def best_of_n_stumps(n):
    """Generate the function to pick the best of n stump evaluations.

    Parameters
    ----------
    n : int
        the number of stumps to evaluate before picking the best. Must be 1
        or more.

    Returns
    -------
    find_best_stump : func
        function to find the best of n stumps.
    """
    if n < 1:
        raise ValueError("n cannot be less than 1")

    def find_best_stump(proximity):
        """Pick the best of n stump evaluations.

        Parameters
        ----------
        proximity : Proximity like object
            the proximity object to split data from.

        Returns
        -------
        stump : ProximityStump
            the best stump / split of data of the n attempts.
        """
        stumps = []
        # for n stumps
        for _ in range(n):
            # duplicate tree configuration
            stump = ProximityStump(
                random_state=proximity.random_state,
                get_exemplars=proximity.get_exemplars,
                distance_measure=proximity.distance_measure,
                setup_distance_measure=proximity.setup_distance_measure,
                get_distance_measure=proximity.get_distance_measure,
                get_gain=proximity.get_gain,
                verbosity=proximity.verbosity,
                n_jobs=proximity.n_jobs,
            )
            # grow the stump
            stump.fit(proximity.X, proximity.y)
            stump.grow()
            stumps.append(stump)
        # pick the best stump based upon gain
        stump = _max(stumps, proximity.random_state, lambda stump: stump.entropy)
        return stump

    return find_best_stump


class ProximityStump(BaseClassifier):
    """Proximity Stump."""

    np.random.seed(1234)

    def __init__(
        self,
        random_state=0,
        setup_distance_measure=setup_all_distance_measure_getter,
        get_distance_measure=None,
        distance_measure=dtw_distance,
        X=None,
        y=None,
        label=None,
        verbosity=0,
        n_stumps=5,
        n_jobs=1,
    ):
        self.setup_distance_measure = setup_distance_measure
        self.random_state = random_state
        self.n_stumps = n_stumps
        self.get_distance_measure = get_distance_measure
        self.distance_measure = distance_measure
        self.verbosity = verbosity
        self.n_jobs = n_jobs
        # set in fit
        self.num_children = None
        self.label_encoder = None
        # exemplars
        self.y_exemplar = None
        self.X_exemplar = None
        # temp_exemplars
        self.temp_exemplar = dict()
        # best_splits
        self.X_best_splits = None
        self.y_best_splits = None
        # Datasets
        self.X = X
        self.y = y
        # splits
        self.children = list()
        self.is_leaf = False
        self.classes_ = dict()
        self.label = label
        self.entropy = None
        super(ProximityStump, self).__init__()

    def set_X(self, X):
        """Set X."""
        self.X = X

    def set_y(self, y):
        """Set y."""
        self.y = y

    @staticmethod
    def gini_gain(y, y_subs):
        """Get gini score of a split, i.e. the gain from parent to children.

        Parameters
        ----------
        y : 1d array like
            array of class labels at parent
        y_subs : list of 1d array like
            list of array of class labels, one array per child

        Returns
        -------
        score : float
            gini score of the split from parent class labels to children. Note a
            higher score means better gain,
            i.e. a better split
        """
        y = np.array(y)
        # find number of instances overall
        parent_n_instances = y.shape[0]
        # if parent has no instances then is pure
        if parent_n_instances == 0:
            for child in y_subs:
                if len(child) > 0:
                    raise ValueError("children populated but parent empty")
            return 0.5
        # find gini for parent node
        score = ProximityStump.gini(y)
        # sum the children's gini scores
        for index in range(len(y_subs)):
            child_class_labels = y_subs[index]
            # ignore empty children
            if len(child_class_labels) > 0:
                # find gini score for this child
                child_score = ProximityStump.gini(child_class_labels)
                # weight score by proportion of instances at child compared to
                # parent
                child_size = len(child_class_labels)
                child_score *= child_size / parent_n_instances
                # add to cumulative sum
                score -= child_score
        return score

    @staticmethod
    def gini(y):
        """Get gini score at a specific node.

        Parameters
        ----------
        y : 1d numpy array
            array of class labels

        Returns
        -------
        score : float
            gini score for the set of class labels (i.e. how pure they are). A
            larger score means more impurity. Zero means
            pure.
        """
        y = np.array(y)
        # get number instances at node
        try:
            n_instances = y.shape[0]
        except exception:
            n_instances = 0

        if n_instances > 0:
            # count each class
            unique_class_labels, class_counts = np.unique(y, return_counts=True)
            # subtract class entropy from current score for each class
            class_counts = np.divide(class_counts, n_instances)
            class_counts = np.power(class_counts, 2)
            sum = np.sum(class_counts)
            return 1 - sum
        else:
            # y is empty, therefore considered pure
            return 0
            # raise ValueError(' y empty')

    @staticmethod
    def split_X_per_class(X, y):
        """Split by class.

        :param X: Array-like containing instances
        :param y: Array-like containing class labels
        :return: Returns a dictionary {Label: [sub_X]} in which sub_X contains all
        instances that match that label
        """
        split_class_x = dict()
        y_size = len(y)
        for index in range(y_size):
            label = y[index]
            if not split_class_x.keys().__contains__(label):
                split_class_x[label] = list()
            if X.shape == 3:
                split_class_x[label].append(X[index][0])
            else:
                split_class_x[label].append(X[index])
        return split_class_x

    @staticmethod
    def calculate_dist_to_exemplars_inst(exemplars, instance, distance_measure):
        """Calculate distance to exemplars."""
        distances = list()
        indices = list()
        if len(exemplars) == 0:
            return None
        for index in range(len(exemplars)):
            exemplar = exemplars[index]
            try:
                distance = distance_measure(instance, exemplar)
            except exception:
                distance = np.inf
            distances.append(distance)
            indices.append(index)
        return distances, indices

    @staticmethod
    def find_closest_distances_inst(exemplars, instance, distance_measure):
        """Find closest distance instance."""
        distances = list()
        indices = list()
        min_distance = np.math.inf
        if (exemplars is None) or len(exemplars) == 0:
            return None, None
        for index in range(len(exemplars)):
            exemplar = exemplars[index][0]
            try:
                distance = distance_measure(instance, exemplar)
            except Exception:
                distance = np.inf
            if len(indices) == 0:
                min_distance = distance
                distances.append(distance)
                indices.append(index)
            else:
                if distance < min_distance:
                    min_distance = distance
                    distances.clear()
                    distances.append(distance)
                    indices.clear()
                    indices.append(index)
                elif distance == min_distance:
                    distances.append(distance)
                    indices.append(index)
        return distances, indices

    @staticmethod
    def find_closest_distance(exemplars, instance, distance_measure):
        """Find closest distance."""
        distance, indices = ProximityStump.find_closest_distances_inst(
            exemplars, instance, distance_measure
        )
        if distance is None:
            return -1, -1
        elif len(distance) == 1:
            return distance[0], indices[0]
        else:
            r = np.random.randint(0, len(distance))
        return distance[r], indices[r]

    def find_closest_distance_(self, instance, distance_measure):
        """Find closest distance."""
        return ProximityStump.find_closest_distance(
            self.X_exemplar, instance, distance_measure
        )

    def find_closest_exemplar_indices(self, X):
        """Find closes exemplar indices."""
        check_X(X)  # todo make checks optional and propogate from forest downwards
        n_instances = X.shape[0]
        distances = self.distance_to_exemplars(X)
        indices = np.empty(X.shape[0], dtype=int)
        for index in range(n_instances):
            exemplar_distances = distances[index]
            closest_exemplar_index = arg_mins(exemplar_distances, self.random_state)
            indices[index] = closest_exemplar_index[0]
        return indices

    def split_stump(self, X, y, dataset_per_class):
        """Split stump."""
        splits_x = dict()  # {index: x_list}
        splits_y = dict()  # {index: y_list}
        label_branch = 0
        for label in dataset_per_class.keys():
            if len(dataset_per_class[label]) == 0:
                continue
            else:
                sub_X = dataset_per_class[
                    label
                ]  # sub_X is a list of series/arrays who belong to a label
                r = np.random.randint(0, len(sub_X))  # select a random element in sub_X
                splits_x[label_branch] = list()
                splits_y[label_branch] = list()
                self.temp_exemplar[label_branch] = sub_X[r]  # sub_X[r] is a serie
                self.classes_[label_branch] = label
                label_branch = label_branch + 1
        for j in range(X.shape[0]):
            instance = self.X[j][0]
            closest_distance, index = ProximityStump.find_closest_distance(
                self.temp_exemplar, instance, self.distance_measure
            )
            if closest_distance == -1:
                return splits_x, splits_y
            splits_x[index].append(X[j][0])
            splits_y[index].append(y[j])
        return splits_x, splits_y  # <index, list_x>, <index, list_y>

    def find_best_stumps(self, X, y):
        """Find best stumps."""
        x_per_label = self.split_X_per_class(X, y)
        best_weighted_gini = np.inf
        x_size = len(X)
        for _ in range(self.n_stumps):
            splits_x, splits_y = self.split_stump(X, y, x_per_label)
            if len(splits_x) == 0:
                return self.X_best_splits, self.y_best_splits
            weighted_gini = self.weighted_gini(x_size, splits_x, splits_y)
            if weighted_gini < best_weighted_gini:
                best_weighted_gini = weighted_gini
                self.X_best_splits = splits_x
                self.y_best_splits = splits_y
                self.X_exemplar = self.temp_exemplar
                self.y_exemplar = self.temp_exemplar.keys()
        self.num_children = len(self.X_best_splits)
        return self.X_best_splits, self.y_best_splits

    @staticmethod
    def weighted_gini(x_size, splits_x, splits_y):
        """Find weighted Gini."""
        wgini = 0.0
        for index in range(len(splits_x)):
            spt_x = splits_x[index]
            spt_y = splits_y[index]
            wgini = wgini + (len(spt_x) / x_size) * ProximityStump.gini(spt_y)
        return wgini

    def calculate_distance_to_exemplars(self, X):
        """Find distance to exemplars.

        :param X: the dataset containing a list of instances
        :return: 2d numpy array of distances from each instance to each
        exemplar (instance by exemplar)
        """
        check_X(X)
        if self.n_jobs > 1 or self.n_jobs < 0:
            parallel = Parallel(self.n_jobs)
            distances = parallel(
                delayed(self.calculate_dist_to_exemplars_inst)(
                    self.X_exemplar, X[0][index, :], self.distance_measure
                )
                for index in range(X.shape[0])
            )
        else:
            distances = [
                self.calculate_dist_to_exemplars_inst(
                    self.X_exemplar, X[index, :][0], self.distance_measure
                )
                for index in range(X.shape[0])
            ]
        return distances

    def distance_to_exemplars(self, X):
        """Distance to exemplars."""
        distances = self.calculate_distance_to_exemplars(X)
        distances = [x[0][0] for x in distances]
        distances = np.vstack(np.array(distances))
        return distances

    def distance_to_exemplars_indices(self, X):
        """Distance to exemplars indices."""
        distances_indices = self.calculate_distance_to_exemplars(X)
        indices = [x[1][0] for x in distances_indices]
        return indices

    def fit(self, X=None, y=None):
        """Fit."""
        if X is None:
            X = self.X
        if y is None:
            y = self.y
        if len(y) == 0:
            return
        gini = ProximityStump.gini(y)
        if gini == 0:
            self.label = int(y[0])
            self.is_leaf = True
            return
        if len(X.shape) == 2:
            X = X.reshape((X.shape[0], 1, X.shape[1]))
        self.find_best_stumps(X, y)
        if len(self.X_best_splits) > 0:
            for i in range(0, len(self.X_best_splits.values())):
                self.children.append(
                    ProximityStump(
                        X=X, y=y, label=y[i], distance_measure=self.distance_measure
                    )
                )
            counter = 0
            for index in self.X_best_splits.keys():
                x_branches = self.X_best_splits[index]
                y_branches = self.y_best_splits[index]
                try:
                    splits_x = np.array(x_branches)
                    splits_y = np.array(y_branches)
                    if splits_x.shape == 2:
                        splits_x = splits_x.reshape(
                            (splits_x.shape[0], 1, splits_x.shape[1])
                        )
                    self.children[counter].fit(splits_x, splits_y)
                except RecursionError:
                    return
                counter = counter + 1

    def predict_proba(self, X):
        """Find probability estimates for each class for all cases in X.

        Parameters
        ----------
        X : array-like or sparse matrix of shape = [n_instances, n_columns]
            The training input samples.
            If a Pandas data frame is passed (sktime format)
            If a Pandas data frame is passed, a check is performed that it
            only has one column.
            If not, an exception is thrown, since this classifier does not
            yet have
            multivariate capability.

        Returns
        -------
        output : array of shape = [n_instances, n_classes] of probabilities
        """
        X = check_X(X, enforce_univariate=True)
        distances = self.distance_to_exemplars(X)
        ones = np.ones(distances.shape)
        distances = np.add(distances, ones)
        distributions = np.divide(ones, distances)
        normalize(distributions, copy=False, norm="l1")
        return distributions


class ProximityTree(BaseClassifier):
    """Proximity Tree class to model a distance based decision tree.

    Attributes
    ----------
        label_encoder: label encoder to change string labels to numeric indices
        classes_: unique list of classes
        random_state: the random state
        get_exemplars: function to extract exemplars from a dataframe and
        class value list
        setup_distance_measure: function to setup the distance measure
        getters from dataframe and class value list
        get_distance_measure: distance measure getters
        distance_measure: distance measures
        get_gain: function to score the quality of a split
        verbosity: logging verbosity
        n_jobs: number of jobs to run in parallel *across threads"
        find_stump: function to find the best split of data
        max_depth: max tree depth
        depth: current depth of tree, as each node is a tree itself,
        therefore can have a depth of >=0
        X: train data
        y: train data labels
        stump: the stump used to split data at this node
        branches: the partitions of data driven by the stump
    """

    def __init__(
        self,
        # note: any changes of these params must be reflected in
        # the fit method for building trees / clones
        random_state=0,
        get_exemplars=get_one_exemplar_per_class_proximity,
        distance_measure=dtw_distance,
        get_distance_measure=None,
        setup_distance_measure=setup_all_distance_measure_getter,
        get_gain=gini_gain,
        max_depth=np.math.inf,
        is_leaf=pure,
        verbosity=0,
        n_jobs=1,
        n_stump_evaluations=5,
        find_stump=None,
    ):
        """Build a Proximity Tree object.

        :param random_state: the random state
        :param get_exemplars: get the exemplars from a given dataframe and
        list of class labels
        :param distance_measure: distance measure to use
        :param get_distance_measure: method to get the distance measure if
        no already set
        :param setup_distance_measure: method to setup the distance measures
        based upon the dataset given
        :param get_gain: method to find the gain of a data split
        :param max_depth: maximum depth of the tree
        :param is_leaf: function to decide when to mark a node as a leaf node
        :param verbosity: number reflecting the verbosity of logging
        :param n_jobs: number of parallel threads to use while building
        :param find_stump: method to find the best split of data / stump at
        a node
        :param n_stump_evaluations: number of stump evaluations to do if
        find_stump method is None
        """
        super().__init__()
        self.verbosity = verbosity
        self.n_stump_evaluations = n_stump_evaluations
        self.find_stump = find_stump
        self.max_depth = max_depth
        self.get_distance_measure = distance_measure
        self.random_state = check_random_state(random_state)
        self.get_distance_measure = get_distance_measure
        self.setup_distance_measure = setup_all_distance_measure_getter
        self.get_gain = get_gain
        self.n_jobs = n_jobs
        self.depth = 0
        # below set in fit method
        self.label_encoder = None
        self.distance_measure = distance_measure
        self.root_stump = None
        self.branches = None
        self.X = None
        self.y = None
        self._is_fitted = False
        self.classes_ = None

    def fit(self, X, y, random_state=0):
        """Fit."""
        self.classes_ = np.unique(y)
        self.root_stump = ProximityStump(
            X=X,
            y=y,
            n_stumps=self.n_stump_evaluations,
            distance_measure=self.distance_measure,
            random_state=self.random_state,
        )
        self.root_stump.fit()
        self._is_fitted = True

    def predict_class_label(self, query):
        """Predict the label of a query.

        :param query:
        :return:
        """
        stump = self.root_stump
        while not stump.is_leaf:
            child_index = stump.find_closest_distance_(query, self.distance_measure)[1]
            if child_index == -1:
                stump.is_leaf = True
                continue
            stump = stump.children[child_index]
        return stump.label

    def predict_proba(self, X):
        """Predict proba."""
        n_instances = X.shape[0]
        predictions = np.zeros(n_instances, dtype=int)
        for index in range(n_instances):
            query = X[index][0]
            class_label = self.predict_class_label(query)
            predictions[index] = class_label
        return predictions


class ProximityForest(BaseClassifier):
    """Proximity Forest class to model a decision tree forest.

    Which uses distance measures to partition data.

    @article{lucas19proximity,

        title={Proximity Forest: an effective and scalable distance-based
        classifier for time series},
        author={B. Lucas and A. Shifaz and C. Pelletier and L. O’Neill and N.
        Zaidi and B. Goethals and F. Petitjean and G. Webb},
         journal={Data Mining and Knowledge Discovery},
        volume={33},
        number={3},
        pages={607--635},
        year={2019}
        }

    Attributes
    ----------
         label_encoder: label encoder to change string labels to numeric indices
         classes_: unique list of classes
         random_state: the random state
         get_exemplars: function to extract exemplars from a dataframe and
         class value list
         setup_distance_measure_getter: function to setup the distance
         measure getters from dataframe and class value list
         get_distance_measure: distance measure getters
         distance_measure: distance measures
         get_gain: function to score the quality of a split
         verbosity: logging verbosity
         n_jobs: number of jobs to run in parallel *across threads"
         find_stump: function to find the best split of data
         max_depth: max tree depth
         X: train data
         y: train data labels
         trees: list of trees in the forest
    """

    def __init__(
        self,
        random_state=0,
        n_estimators=100,
        distance_measure=dtw_distance,
        get_distance_measure=None,
        get_exemplars=get_one_exemplar_per_class_proximity,
        get_gain=gini_gain,
        verbosity=0,
        max_depth=np.math.inf,
        is_leaf=pure,
        n_jobs=1,
        n_stump_evaluations=5,
        find_stump=None,
        setup_distance_measure_getter=setup_all_distance_measure_getter,
    ):
        """Build a Proximity Forest object.

        :param random_state: the random state
        :param get_exemplars: get the exemplars from a given dataframe and
        list of class labels
        :param distance_measure: distance measure to use
        :param get_distance_measure: method to get the distance measure if
        no already set
        :param setup_distance_measure_getter: method to setup the distance
        measures based upon the dataset given
        :param get_gain: method to find the gain of a data split
        :param max_depth: maximum depth of the tree
        :param is_leaf: function to decide when to mark a node as a leaf node
        :param verbosity: number reflecting the verbosity of logging
        :param n_jobs: number of parallel threads to use while building
        :param find_stump: method to find the best split of data / stump at
        a node
        :param n_stump_evaluations: number of stump evaluations to do if
        find_stump method is None
        :param n_estimators: number of trees to construct
        """
        self.verbosity = verbosity
        self.max_depth = max_depth
        self.get_exemplars = get_exemplars
        self.get_gain = get_gain
        self.random_state = random_state
        self.n_estimators = n_estimators
        self.n_jobs = n_jobs
        self.n_stump_evaluations = n_stump_evaluations
        self.get_distance_measure = get_distance_measure
        self.setup_distance_measure_getter = setup_distance_measure_getter
        self.distance_measure = distance_measure
        self.find_stump = find_stump
        # set in fit method
        self.label_encoder = None
        self.X = None
        self.y = None
        self.classes_ = None
        self.trees = list()
        self.num_classes_predicted = dict()

        for _ in range(self.n_estimators):
            self.trees.append(
                ProximityTree(
                    n_stump_evaluations=self.n_stump_evaluations,
                    distance_measure=self.distance_measure,
                )
            )
        super(ProximityForest, self).__init__()

    def _fit_tree(self, X, y, index, random_state=0):
        self.trees[index].fit(X, y, random_state)
        return self.trees[index]

    @staticmethod
    def _predict_proba_tree(X, tree):
        return tree.predict_proba(X)

    def fit_tree(self, X, y, index, random_state):
        """Build the classifier on the training set (X, y).

        X : array-like or sparse matrix of shape = [n_instances,n_columns]
            The training input samples.  If a Pandas data frame is passed,
            column 0 is extracted.
        y : array-like, shape = [n_instances]
            The class labels.
        index : index of the tree to be constructed
        random_state: random_state to send to the tree to be constructed

        Returns
        -------
        self : object
        """
        if self.verbosity > 0:
            print("tree " + str(index) + " building")  # noqa
        tree = ProximityTree(
            random_state=random_state,
            verbosity=self.verbosity,
            distance_measure=self.distance_measure,
            get_distance_measure=self.get_distance_measure,
            max_depth=self.max_depth,
            n_jobs=1,
            find_stump=self.find_stump,
            n_stump_evaluations=self.n_stump_evaluations,
        )
        tree.fit(X, y)
        return tree

    def fit(self, X, y):
        """Fit."""
        X, y = check_X_y(X, y, enforce_univariate=True)
        X = from_nested_to_3d_numpy(X)
        self.random_state = check_random_state(self.random_state)
        # setup label encoding
        if self.label_encoder is None:
            self.label_encoder = LabelEncoder()
            y = self.label_encoder.fit_transform(y)
        self.y = y
        self.classes_ = self.label_encoder.classes_
        if self.distance_measure is None:
            if self.get_distance_measure is None:
                self.get_distance_measure = self.setup_distance_measure_getter(self)
            self.distance_measure = self.get_distance_measure(self)
        if self.n_jobs > 1 or self.n_jobs < 0:
            parallel = Parallel(self.n_jobs)
            self.trees = parallel(
                delayed(self._fit_tree)(
                    X, y, index, self.random_state.randint(0, self.n_estimators)
                )
                for index in range(self.n_estimators)
            )
        else:
            self.trees = [
                self._fit_tree(
                    X, y, index, self.random_state.randint(0, self.n_estimators)
                )
                for index in range(self.n_estimators)
            ]
        self._is_fitted = True
        return self

    def predict_proba(self, X):
        """Find probability estimates for each class for all cases in X.

        Parameters
        ----------
        X : array-like or sparse matrix of shape = [n_instances, n_columns]
            The training input samples.
            If a Pandas data frame is passed (sktime format)
            If a Pandas data frame is passed, a check is performed that it
            only has one column.
            If not, an exception is thrown, since this classifier does not
            yet have
            multivariate capability.

        Returns
        -------
        output : array of shape = [n_instances, n_classes] of probabilities
        """
        X = from_nested_to_3d_numpy(X)
        X = check_X(X, enforce_univariate=True)
        if self.n_jobs > 1 or self.n_jobs < 0:
            parallel = Parallel(self.n_jobs)
            predictions_per_tree = parallel(
                delayed(self._predict_proba_tree)(X, tree) for tree in self.trees
            )
        else:
            predictions_per_tree = [
                self._predict_proba_tree(X, tree) for tree in self.trees
            ]
        distributions = self.calculate_distributions(predictions_per_tree, X.shape[0])
        distributions = np.array(distributions)
        normalize(distributions, copy=False, norm="l1")
        return distributions

    def calculate_prediction_counts(self, predictions):
        """Pick an array of labels predicted by the trees.

        Reorganize it into a dictionary {label: times predicted}
        :param predictions: Array-like which contains the label predicted by each tree
        :return:
        """
        arr = np.array(predictions)
        count_arr = np.bincount(arr)
        prediction_counts = dict()
        for label in np.sort(np.unique(self.y)):
            try:
                prediction_counts[label] = count_arr[label]
            except IndexError:
                prediction_counts[label] = 0
        return prediction_counts

    def calculate_distributions(self, predictions_per_tree, size):
        """Find probability estimates for each class for all cases in X.

        :param predictions_per_tree: Array-like of shape
        [n_instances,[n_tree_estimators,labels]] which contains an array of labels
        predicted by each tree for each instance.
        :param size: Size of X dataset containing the instances to predict
        :return:
        """
        distributions = np.zeros((size, len(np.unique(self.classes_))))
        for index in range(size):
            predicted_classes = list()
            for predict_tree in predictions_per_tree:
                predicted_classes.append(predict_tree[index])
            prediction_counts = self.calculate_prediction_counts(predicted_classes)
            prediction_counts = list(prediction_counts.items())
            prediction_counts_to_array = np.array(prediction_counts)
            sub_distribution = prediction_counts_to_array[
                np.argsort(prediction_counts_to_array[:, 0])
            ][:, 1]
            np.add.at(distributions, index, sub_distribution)
        return distributions
